

import dis
import inspect

from pycarol.pipeline.utils import (
    int_to_bytes,
    flat_list,
    is_builtin,
    is_function,
    is_code,
    enumerate_with_context
)

#TODO: move this module to a dedicated repository
DEBUG_MODE = False

from pycarol.pipeline.utils.hash_versioning.inspect_bytecode import (
    get_name_and_code_of_MAKE_FUNCTION,
    get_name_and_object_of_IMPORT_NAME,
    get_name_and_object_of_CALL_FUNCTION,
    get_name_and_object_of_LOAD_ATTR,
)


def _get_consts_hash(f) -> bytes:
    """Returns hash code of local consts"""
    if hasattr(f, '__code__'):  # is function object
        consts_list = list(f.__code__.co_consts)
    elif hasattr(f, 'co_code'):  # is code object
        consts_list = list(f.co_consts)
    else:
        raise TypeError

    for i, v in enumerate(consts_list):
        if isinstance(v, str) and "<locals>" in v:
            consts_list[i] = v.split('.')[-1]
    consts_tuple = tuple(consts_list)
    return int_to_bytes(hash(consts_tuple))

def process_op(
        context,
        defined_functions,
        code_set,
        called_functions,
        parent_function,
        robust
        ):
    """
    Process instruction updating defined_functions, code_set and
    called_functions accordingly
    Args:
        context: ix, inst, instructions. bytecode itself
        defined_functions: set of functions defined until now
        code_set: set of functions already inspected for hash versioning purpose
        called_functions: set of called functions until now in this bytecode
        parent_function: function object of this bytecode
        robust: set true to ignore internal assertion errors

    Returns:

    """
    ix, inst, instructions = context
    if inst.opname == "MAKE_FUNCTION":
        defined_functions.update(
            get_name_and_code_of_MAKE_FUNCTION(*context))

    if inst.opname == "IMPORT_NAME": #TODO: some tests are failing for IMPORT_NAME
        defined_functions.update(
            get_name_and_object_of_IMPORT_NAME(*context))

    if "CALL_FUNCTION" in inst.opname:
        try:
            _, son_function = get_name_and_object_of_CALL_FUNCTION(
                ix,inst,instructions,
                parent_function,
                defined_functions
            )
        except AssertionError as error:
            if robust:
                # print(error)
                return  # do not update code_set and called_functions
            else:
                raise(error)

        if son_function not in code_set:
            code_set.add(son_function)
            called_functions.add(son_function)
    return

def traverse_code(parent_function: 'function', code_set,
                  ignore_not_implemented=True) -> list:
    """
    This method recursively traverse the bytecodes of top_function entering
    in every function call found. At every level it returns a list of
    bytecodes of respective functions called.

    Args:
        parent_function: function to be analyzed at this recursion level
        code_set: set of functions already analyzed
        ignore_not_implemented: if True, will generate bytecode_tree where it is
        possible. Where it is not possible, will return empty list.

    Returns:
        bytecode_tree: nested lists of bytecode

    """

    if is_builtin(parent_function):
        # dis cannot get instructions for builtins
        # return function name instead of bytecode
        return [parent_function.__name__]

    called_functions = set()
    defined_functions: dict = {}

    # linear search on bytecodes for called functions
    instructions = list(dis.get_instructions(parent_function))
    for context in enumerate_with_context(instructions):
        try:
            process_op(context,defined_functions,code_set,called_functions,
                       parent_function,ignore_not_implemented)
        except NotImplementedError as e:
            if ignore_not_implemented:
                continue
            else:
                raise e

    # recursion step
    code_list = [traverse_code(f, code_set, ignore_not_implemented) for f in
                 called_functions]

    # leaf node step
    if is_function(parent_function):

        function_code: list = b''.join([
            parent_function.__code__.co_code,
            _get_consts_hash(parent_function),
            int_to_bytes(hash(
                dict(inspect.getmembers(parent_function)).get('__defaults__',0)
            )),
        ])
    elif is_code(parent_function):  # parent_function is code object
        function_code: list = b''.join([
            parent_function.co_code,
            _get_consts_hash(parent_function),
            # missing default parameter information
            # asbytes(hash(
            #     dict(inspect.getmembers(parent_function))['__defaults__']
            # )),
        ])
    else:
        raise TypeError("parent_function should be either function or code.")

    code_list.append(function_code)
    return code_list


def get_bytecode_tree(
        top_function: 'function',
        ignore_not_implemented=False,
) -> list:
    """
    Args:
        top_function: analyzed function
        ignore_not_implemented: if True, will generate bytecode_tree where it is
        possible. Where it is not possible, will return empty list.

    Returns:
        bytecode_tree: nested lists of bytecode

    """
    #TODO: nested, builtin, call_kwargs are failing interprocesses test
    from functools import partial
    if isinstance(top_function,partial):
        top_function = top_function.func

    assert is_function(top_function), 'argument should be a function'
    code_set = set()  # "global" memoization set
    bytecode_tree = traverse_code(top_function, code_set,ignore_not_implemented)

    return bytecode_tree

def get_function_hash(f: 'function', ignore_not_implemented=False) -> int:
    """
    Module main function. It returns a proper hash for the given function.
    Args:
        f: function to be hashed
        ignore_not_implemented: if True, will consider bytecode_tree where it is
        possible. Where it is not possible, will replace by an empty list.
        and compute a hash anyway. Some part of the code may be ignored.

    Returns:
        h: a hash number for the given function

    """
    bytecode_nested_list = get_bytecode_tree(f, ignore_not_implemented)
    bytecode_flat_list = flat_list(bytecode_nested_list)
    bytecode_flat_list = [
        bytecode.encode()
        if isinstance(bytecode,str)
        else bytecode
        for bytecode in bytecode_flat_list
    ]
    
    from hashlib import sha256
    hashable_bytecode = b''.join(bytecode_flat_list)
    h = sha256(hashable_bytecode).hexdigest()
    return h





import json

class ManifestParser():

    def __init__(self, manifest_json):
        self.manifest_json = manifest_json


    def has_batch_process(self):
        return 'batch' in self.manifest_json and self.manifest_json['batch'] != {}


    def has_batch_process_name(self, name):
        if self.has_batch_process():
            for process in self.manifest_json['batch']['processes']:
                if process['name'] == name:
                    return process
        return None


    def has_online_process(self):
        return 'online' in self.manifest_json and self.manifest_json['online'] != {}


    def has_online_process_name(self, name):
        if self.has_online_process():
            for process in self.manifest_json['online']['processes']:
                if process['name'] == name:
                    return process
        return None


    def has_instance_properties_batch(self, name):
        process = self.has_batch_process_name(name)
        if process:
            if 'instanceProperties' in process and process['instanceProperties'] != {}:
                return True
        return False


    def chosen_profile_batch(self, name):
        process = self.has_batch_process_name(name)
        if process:
            chosenProfile = process['instanceProperties']['profile']
            return 'default' if chosenProfile == '' else chosenProfile
        return None


    def keep_instance_batch(self, name):
        process = self.has_batch_process_name(name)
        if process and 'keepInstance' in process['instanceProperties']:
            keepInstance = process['instanceProperties']['keepInstance']
            return True if keepInstance == 'True' or keepInstance == 'true' else False
        return False


    def docker_image_batch(self, name):
        process = self.has_batch_process_name(name)
        if process:
            return process['instanceProperties']['properties']['dockerImage']
        return None


    def instance_memory_batch(self, name):
        process = self.has_batch_process_name(name)
        if process:
            return process['instanceProperties']['properties']['instanceMemory']
        return None


    def instance_so_batch(self, name):
        process = self.has_batch_process_name(name)
        if process:
            return process['instanceProperties']['properties']['instanceSO']
        return None


    def instance_size_batch(self, name):
        process = self.has_batch_process_name(name)
        if process:
            return process['instanceProperties']['properties']['instanceSize']
        return None


    def instance_vcpus_batch(self, name):
        process = self.has_batch_process_name(name)
        if process:
            return process['instanceProperties']['properties']['instanceVCPUs']
        return None


    def luigi_workers_batch(self, name):
        process = self.has_batch_process_name(name)
        if process and 'luigi' in process['instanceProperties']:
            if process and 'workers' in process['instanceProperties']['luigi']:
                return process['instanceProperties']['luigi']['workers']
        return '2'


    def instance_environment_variables_batch(self, name):
        process = self.has_batch_process_name(name)
        if process and 'environments' in process['instanceProperties']:
            return process['instanceProperties']['environments']
        return None


    def algorithm_description_batch(self, name):
        process = self.has_batch_process_name(name)
        if process:
            if 'algorithmDescription' in process:
                if 'pt-br' in process:
                    return process['algorithmDescription']['pt-br']
                elif 'en-us' in process:
                    return process['algorithmDescription']['en-us']
        return ''


    def algorithm_name_batch(self, name):
        process = self.has_batch_process_name(name)
        if process:
            return process['algorithmName']
        return None


    def algorithm_method_batch(self, name):
        process = self.has_batch_process_name(name)
        if process:
            return process['algorithmMethod']
        return None


    def algorithm_title_batch(self, name):
        process = self.has_batch_process_name(name)
        if process:
            if 'algorithmTitle' in process:
                if 'pt-br' in process:
                    return process['algorithmTitle']['pt-br']
                elif 'en-us' in process:
                    return process['algorithmTitle']['en-us']
        return ''


    def git_batch(self, name):
        process = self.has_batch_process_name(name)
        if process and 'git' in process:
            return process['git']
        return None


    def has_instance_properties_online(self, name):
        process = self.has_online_process_name(name)
        if process:
            if 'instanceProperties' in process and process['instanceProperties'] != {}:
                return True
        return False


    def docker_image_online(self, name):
        process = self.has_online_process_name(name)
        if process:
            return process['instanceProperties']['properties']['dockerImage']


    def instance_memory_online(self, name):
        process = self.has_online_process_name(name)
        if process:
            return process['instanceProperties']['properties']['instanceMemory']


    def instance_so_online(self, name):
        process = self.has_online_process_name(name)
        if process:
            return process['instanceProperties']['properties']['instanceSO']


    def instance_size_online(self, name):
        process = self.has_online_process_name(name)
        if process:
            return process['instanceProperties']['properties']['instanceSize']
        return None


    def instance_vcpus_online(self, name):
        process = self.has_online_process_name(name)
        if process:
            return process['instanceProperties']['properties']['instanceVCPUs']
        return None


    def luigi_workers_online(self, name):
        process = self.has_online_process_name(name)
        if process and 'luigi' in process['instanceProperties']:
            return process['instanceProperties']['luigi']['workers']
        return None


    def instance_environment_variables_online(self, name):
        process = self.has_online_process_name(name)
        if process and 'environments' in process['instanceProperties']:
            return process['instanceProperties']['environments']
        return None


    def algorithm_description_online(self, name):
        process = self.has_online_process_name(name)
        if process:
            if 'algorithmDescription' in process:
                if 'pt-br' in process:
                    return process['algorithmDescription']['pt-br']
                elif 'en-us' in process:
                    return process['algorithmDescription']['en-us']
        return ''


    def name_online(self, name):
        process = self.has_online_process_name(name)
        if process:
            return process['name']
        return None


    def algorithm_name_online(self, name):
        process = self.has_online_process_name(name)
        if process:
            return process['algorithmName']
        return None


    def algorithm_title_online(self, name):
        process = self.has_online_process_name(name)
        if process:
            if 'algorithmTitle' in process:
                if 'pt-br' in process:
                    return process['algorithmTitle']['pt-br']
                elif 'en-us' in process:
                    return process['algorithmTitle']['en-us']
        return ''


    def git_online(self, name):
        process = self.has_online_process_name(name)
        if process and 'git' in process:
            return process['git']
        return None


from pathlib import Path

from dotenv import load_dotenv
import pandas as pd

import pycarol

TEST_FILEPATH = Path("./test/artifacts/test.txt")
TEST_FILENAME = TEST_FILEPATH.name


def _setup_storage() -> pycarol.Storage:
    load_dotenv()
    login = pycarol.Carol()
    storage = pycarol.Storage(login)
    return storage


def test_save_parquet():
    storage = _setup_storage()
    empty_df = pd.DataFrame()
    storage.save(name=TEST_FILENAME, obj=empty_df, format="pickle")

    back_df = storage.load(TEST_FILENAME, format="pickle")
    assert back_df.empty, "Loaded DataFrame is different from saved one."


def test_save_joblib():
    storage = _setup_storage()
    storage.save(name=TEST_FILENAME, obj=str(TEST_FILEPATH), format="joblib")

    backpath = storage.load(TEST_FILENAME, format="joblib")
    with open(str(TEST_FILEPATH), "r") as file1:
        text1 = file1.read()

    with open(str(backpath), "r") as file2:
        text2 = file2.read()

    assert text1 == text2, "Texts are equal."


def test_save_pickle():
    storage = _setup_storage()
    storage.save(name=TEST_FILENAME, obj=str(TEST_FILEPATH), format="pickle")

    backpath = storage.load(TEST_FILENAME, format="pickle")
    with open(str(TEST_FILEPATH), "r") as file1:
        text1 = file1.read()

    with open(str(backpath), "r") as file2:
        text2 = file2.read()

    assert text1 == text2, "Texts are equal."


def test_save_file():
    storage = _setup_storage()
    storage.save(name=TEST_FILENAME, obj=str(TEST_FILEPATH), format="file")

    backpath = storage.load(TEST_FILENAME, format="file")
    with open(str(TEST_FILEPATH), "r") as file1:
        text1 = file1.read()

    with open(str(backpath), "r") as file2:
        text2 = file2.read()

    assert text1 == text2, "Texts are equal."


def test_save_other():
    storage = _setup_storage()

    try:
        storage.save(name=TEST_FILENAME, obj=str(TEST_FILEPATH), format="another")
    except ValueError:
        return

    raise Exception("Saving with unexistant format should raise exception.")


if __name__ == "__main__":
    test_save_parquet()



from . import *

dag={
    0: [1,2],
    1: [2,3],
    2: [4],
    3: [4],
    4: []
}
l=[]


def test_breadth_first_search():
    def print_visiting_order(node, visiting_list):
        print(visiting_list,'.',node)
        visiting_list.append(node)

    for i,levels in enumerate(
            breadth_first_search(dag,[0],print_visiting_order,l)):
        print(i,levels)


def test_get_dag_node_level():
    levels = get_dag_node_level(dag)
    print(levels)


def test_get_reverse_dag():
    print(dag)
    print(get_reverse_dag(dag))
    print(get_reverse_dag(get_reverse_dag(dag)))
    assert dag == get_reverse_dag(get_reverse_dag(dag))



import os

from pycarol.tasks import Tasks
from pycarol.carol import Carol


class HealthCheckOnline():

    def __init__(self, logs):
        self.domain = os.environ.get('CAROLDOMAIN')
        self.app_name = os.environ.get('CAROLAPPNAME')
        self.app_version = os.environ.get('CAROLAPPVERSION')
        self.online_name = os.environ.get('CAROLONLINENAME')
        self.connector_id = os.environ.get('CAROLCONNECTORID')
        self.app_oauth = os.environ.get('CAROLAPPOAUTH')
        self.long_task_id = os.environ.get('LONGTASKID')

        self.logs = logs


    def carol_auth(self):
        try:
            carol = Carol()
            # pyCarol should return a error if the credentials
            # are invalid, so for now we try to validate if
            # those are valid by another method
            carol.api_key_details(self.app_oauth,self.connector_id)
            return carol
        except Exception as e:
            raise Exception("Carol authentication failed.", e)


    def send_status_carol(self):
        if self.domain:
            carol = self.carol_auth()
            tasks = Tasks(carol)
            task = tasks.get_task(task_id=self.long_task_id)

            if self.logs == []:
                task.set_progress(80)
            else:
                task.set_progress(50)
                task.add_log(str(self.logs), log_level='ERROR')



import json
import typing as T

from dotenv import load_dotenv
import pandas as pd

import pycarol

SA_FILEPATH = "/home/jro/wk/totvs/sa.json"
TEST_QUERY1 = """
    SELECT *
    FROM `4c2c9090e7c611e893bf0e900682978b.dm_clockinrecords`
    LIMIT 100
"""


def _setup_service_account() -> T.Dict[str, str]:
    with open(SA_FILEPATH, "r") as file:
        service_account = json.loads(file.read())
    return service_account


def _setup_login() -> pycarol.Carol:
    load_dotenv()
    return pycarol.Carol()


def test_query():
    carol = _setup_login()
    service_account = _setup_service_account()
    result = pycarol.bigquery.query(carol, TEST_QUERY1, service_account).to_dataframe()
    assert isinstance(result, pd.DataFrame), "result must be a Pandas DataFrame."
    assert not result.empty, "result must not be empty."


def test_staging_pure_query():
    carol = _setup_login()
    service_account = _setup_service_account()
    sql = pycarol.SQL(carol)
    result = sql.query(TEST_QUERY1, method="bigquery", service_account=service_account)
    assert isinstance(result, pd.DataFrame), "result must be a Pandas DataFrame."
    assert not result.empty, "result must not be empty."


def test_staging_pure_query_dict():
    carol = _setup_login()
    service_account = _setup_service_account()
    sql = pycarol.SQL(carol)
    result = sql.query(
        TEST_QUERY1, method="bigquery", service_account=service_account, dataframe=False
    )
    assert isinstance(result, T.List), "result must be a list."
    assert len(result) > 1, "result must not be empty."


def test_staging_templated_query():
    test_query = """
        SELECT *
        FROM {{clockinmobile.clockinrecords}}
        LIMIT 100
    """
    carol = _setup_login()
    service_account = _setup_service_account()
    sql = pycarol.SQL(carol)
    result = sql.query(test_query, method="bigquery", service_account=service_account)
    assert isinstance(result, pd.DataFrame), "result must be a Pandas DataFrame."
    assert not result.empty, "result must not be empty."


def test_model_templated_query():
    test_query = """
        SELECT *
        FROM {{clockinrecords}}
        LIMIT 100
    """
    carol = _setup_login()
    service_account = _setup_service_account()
    sql = pycarol.SQL(carol)
    result = sql.query(test_query, method="bigquery", service_account=service_account)
    assert isinstance(result, pd.DataFrame), "result must be a Pandas DataFrame."
    assert not result.empty, "result must not be empty."


def test_mix_templated_query():
    test_query = """
        SELECT mdmId
        FROM {{clockinmobile.clockinrecords}}
        UNION ALL
        SELECT mdmId
        FROM {{clockinrecords}}
        LIMIT 100
    """
    carol = _setup_login()
    service_account = _setup_service_account()
    sql = pycarol.SQL(carol)
    result = sql.query(test_query, method="bigquery", service_account=service_account)
    assert isinstance(result, pd.DataFrame), "result must be a Pandas DataFrame."
    assert not result.empty, "result must not be empty."


if __name__ == "__main__":
    test_query()
    test_staging_pure_query()
    test_staging_pure_query_dict()
    test_staging_templated_query()
    test_model_templated_query()
    test_mix_templated_query()



import json

import time
import copy
import warnings
import asyncio

from .data_models_fields import DataModelFields
from .data_model_types import DataModelTypeIds

from ..utils.importers import _import_dask, _import_pandas
from ..verticals import Verticals
from ..storage import Storage
from ..query import delete_golden
from ..connectors import Connectors
from ..utils import async_helpers
from ..utils.miscellaneous import stream_data
from .. import _CAROL_METADATA_GOLDEN, _NEEDED_FOR_MERGE, _REJECTED_DM_COLS, _CAROL_METADATA_UNTIE_GOLDEN
from ..utils.miscellaneous import drop_duplicated_parquet, drop_duplicated_parquet_dask
from ..utils.deprecation_msgs import _deprecation_msgs
from ..exceptions import CarolApiResponseException

_DATA_MODEL_TYPES_MAPPING = {
    "boolean": bool,
    "date": str,  # TODO should it be pd.datetime?
    "long": int,
    "double": float,
    "nested": str,
    "string": str,
    "binary": str,
    "enum": str,
    "object": str,
    "geopoint": str
}


class DataModel:

    def __init__(self, carol):
        self.carol = carol
        self.fields_dict = {}
        self.entity_template_ = {}

    def get_staging_mapped(self, dm_name=None, dm_id=None):
        """
        Get mapped stagings to this data model.

        Args:
            dm_name: `str` default `None`
                Datamodel name
            dm_id: `str` default `None`
                Datamodel id

        Returns: `dict`
            A dictionary formated as:
            {
                "connector_id_1" : ["staging1", "staging2"],
                "connector_id_2" : ["staging_3", "staging_4"]
            }
        """

        if dm_name is not None:
            resp = self.get_by_name(dm_name)
            dm_id = resp['mdmId']
        elif dm_id is None:
            raise ValueError('Either `dm_name` or `dm_id` must be set.')
        url = f"v1/entities/templates/{dm_id}/mappedStagingTypes"
        query = self.carol.call_api(url, method='GET')
        return query

    def _build_query_params(self):
        if self.sort_by is None:
            self.query_params = {"offset": self.offset, "pageSize": str(
                self.page_size), "sortOrder": self.sort_order}
        else:
            self.query_params = {"offset": self.offset, "pageSize": str(self.page_size), "sortOrder": self.sort_order,
                                 "sortBy": self.sort_by}

    def _get_name_type_data_models(self, fields):
        f = {}
        for field in fields:
            if field.get('mdmMappingDataType', None) not in ['NESTED', 'OBJECT']:
                f.update({field['mdmName']: field['mdmMappingDataType']})
            else:
                f[field['mdmName']] = self._get_name_type_data_models(
                    field['mdmFields'])
        return f

    def _get(self, id, by='id'):

        if by == 'name':
            url = f"v1/entities/templates/name/{id}"
        elif by == 'id':
            url = f"v1/entities/templates/{id}/working"
        else:
            raise print('Type incorrect, it should be "id" or "name"')

        # TODO: Add 'Not Found' and `is in Deleted state`
        resp = self.carol.call_api(url, method='GET')
        self.entity_template_ = {resp['mdmName']: resp}
        self.fields_dict.update(
            {resp['mdmName']: self._get_name_type_data_models(resp['mdmFields'])})
        return resp

    def fetch_parquet(
            self, dm_name, merge_records=True, backend='pandas',
            return_dask_graph=False,
            columns=None, return_metadata=False, callback=None,
            max_hits=None, cds=True, max_workers=None, file_pattern=None,
            return_callback_result=False
    ):
        """
        Fetch parquet from Golden.

        Args:
            dm_name: `str`
                Data model name to be imported
            merge_records: `bool`, default `True`
                This will keep only the most recent record exported. Sometimes there are updates and/or deletions and
                one should keep only the last records.
            backend: ['dask','pandas'], default `dask`
                if to use either dask or pandas to fetch the data
            return_dask_graph: `bool`, default `false`
                If to return the dask graph or the dataframe.
            columns: `list`, default `None`
                List of columns to fetch.
            return_metadata: `bool`, default `False`
                To return or not the fields like ['mdmId', 'mdmCounterForEntity', etc.]
            callback: `callable`, default `None`
                Function to be called each downloaded file.
            max_hits: `int`, default `None`
                Number of records to get.
            cds: `bool`, default `False`
                Get records from CDS.
            max_workers: `int` default `None`
                Number of workers to use when downloading parquet files with pandas back-end.
            file_pattern: `str` default `None`
                File pattern to filter data when fetching from CDS. e.g.
                file_pattern='2019-11-25' will fetch only CDS files that start with `2019-11-25`.
            return_callback_result `bool` default `False`
                If a callback is used, it will return the result of the response of the callback. This will skip all the
                operation to merge records and return selected columns.
            :return:
            """

        if backend not in ('dask', 'pandas'):
            raise ValueError(
                f"Backend options are 'dask','pandas' {backend} was given")

        if return_metadata:
            # It can be costly to get all meta from a golden. So er should alway ask for the info we want.
            _meta_cols = _CAROL_METADATA_GOLDEN
        else:
            _meta_cols = _NEEDED_FOR_MERGE

        if callback and not callable(callback):
            raise TypeError(f'"{callback}" object is not callable')

        all_cols = list(self._get_name_type_DMs(
            self.get_by_name(dm_name)['mdmFields']))
        if not columns:  # if an empty list was sent.
            columns = all_cols

        elif isinstance(columns, str):
            columns = [columns]

        _diff_cols = set(columns) - set(all_cols)
        if len(_diff_cols) > 0:
            warnings.warn(
                f"It seems there was used columns not in this data model: {_diff_cols}", UserWarning)

        if return_dask_graph and backend != 'dask':
            warnings.warn(
                '`return_dask_graph` has no use when `backend!=dask`')

        if not cds:
            _deprecation_msgs("`cds` option will be removed from pycarol 3.33. Consider use `cds=True`"
                              " to avoid problems. ")

        import_type = 'golden_cds'
        if columns:
            columns.extend(_meta_cols)

        storage = Storage(self.carol)
        token_carolina = storage.backend.carolina.token
        storage_space = storage.backend.carolina.get_bucket_name(import_type)

        if backend == 'dask':
            d = _import_dask(storage=storage, dm_name=dm_name,
                             import_type=import_type,
                             merge_records=merge_records,
                             return_dask_graph=return_dask_graph,
                             columns=columns)

        elif backend == 'pandas':
            import pandas as pd
            d = _import_pandas(storage=storage, dm_name=dm_name,
                               import_type=import_type, columns=columns,
                               callback=callback, max_hits=max_hits,
                               max_workers=max_workers,
                               token_carolina=token_carolina,
                               storage_space=storage_space, file_pattern=file_pattern)
            if d is None:
                warnings.warn("No data to fetch!", UserWarning)
                _field_types = self._get_name_type_DMs(
                    self.get_by_name(dm_name)['mdmFields'])
                cols_keys = list(_field_types)
                if return_metadata:
                    cols_keys.extend(_meta_cols)

                elif columns:
                    columns = [i for i in columns if i not in _meta_cols]

                d = pd.DataFrame(columns=cols_keys)
                for key, value in _field_types.items():
                    if isinstance(value, dict):
                        value = "STRING"  # If nested we receive as a `STR`
                    d.loc[:, key] = d.loc[:, key].astype(
                        _DATA_MODEL_TYPES_MAPPING.get(value.lower(), str), copy=False)
                if columns:
                    columns = list(set(columns))
                    d = d[list(set(columns))]
                return d

        else:
            raise ValueError(
                f'backend should be either "dask" or "pandas" you entered {backend}')

        if (return_callback_result) and (callback is not None):
            return d

        if merge_records:
            if (not return_dask_graph) or (backend == 'pandas'):
                d = drop_duplicated_parquet(
                    d, untie_field=_CAROL_METADATA_UNTIE_GOLDEN)
            else:
                d = drop_duplicated_parquet_dask(
                    d, untie_field=_CAROL_METADATA_UNTIE_GOLDEN)

        if not return_metadata:
            to_drop = set(_meta_cols).intersection(set(d.columns))
            d = d.drop(labels=to_drop, axis=1)

        return d

    def get_all(self, offset=0, page_size=-1, sort_order='ASC',
                sort_by=None, print_status=False,
                save_file=None, only_published=False):
        """ 
        Fetch all datamodels definition. 

        Args:

            offset: `int`, default 0
                Offset for pagination. Only used when `scrollable=False`
            page_size: `int`, default 100
                Number of records downloaded in each pagination. The maximum value is 1000
            sort_order: `str`, default 'ASC'
                Sort ascending ('ASC') vs. descending ('DESC').
            sort_by: `str`,  default `None`
                Name to sort by.
            print_status: `bool`, default False
                Print number of records donwloaded. 
            save_file: `str`, default `None`
                If not `None` the results will be saved. 
            only_published: `bool`, default `False`
                Get only published data models. 


        """

        self.offset = offset
        self.page_size = page_size
        self.sort_order = sort_order
        self.sort_by = sort_by
        self._build_query_params()

        self.template_dict = {}
        self.template_data = []
        count = self.offset

        set_param = True
        self.total_hits = float("inf")
        if save_file:
            assert isinstance(save_file, str)
            file = open(save_file, 'w', encoding='utf8')

        if only_published:
            url_filter = "v1/entities/templates/published"
        else:
            url_filter = "v1/entities/templates"

        while count < self.total_hits:

            query = self.carol.call_api(
                url_filter, params=self.query_params, method='GET')

            if query['count'] == 0:
                print('There are no more results.')
                print('Expecting {}, reponse = {}'.format(
                    self.total_hits, count))
                break
            count += query['count']
            if set_param:
                self.total_hits = query["totalHits"]
                set_param = False

            query = query['hits']
            self.template_data.extend(query)
            self.fields_dict.update({i['mdmName']: self._get_name_type_data_models(i['mdmFields'])
                                     for i in query})
            self.template_dict.update({i['mdmName']: {'mdmId': i['mdmId'],
                                                      'mdmEntitySpace': i['mdmEntitySpace'],
                                                      'mdmPublishedExists': i['mdmPublishedExists']}
                                       for i in query})

            self.query_params['offset'] = count
            if print_status:
                print('{}/{}'.format(count, self.total_hits), end='\r')
            if save_file:
                file.write(json.dumps(query, ensure_ascii=False))
                file.write('\n')
                file.flush()
        if save_file:
            file.close()
        return self

    def get_by_name(self, name):
        return self._get(name, by='name')

    def get_by_id(self, id):
        return self._get(id, by='id')

    def get_snapshot(self, dm_id, entity_space):
        url_snapshot = f'v1/entities/templates/{dm_id}/snapshot?entitySpace={entity_space}'
        resp = self.carol.call_api(url_snapshot, method='GET')
        self.snapshot_ = {resp['entityTemplateName']: resp}
        return resp

    def export(self, dm_name=None, dm_id=None, sync_dm=True, full_export=False, delete_previous=False):
        """

        @DEPRECATED. This function was removed in pycarol 3.34

        Export data models

        This method will trigger or pause the export of the data in the datamodel to
        CDS

        :param dm_name: `str`, default `None`
            Data model Name
        :param dm_id: `str`, default `None`
            Data model id
        :param sync_dm: `bool`, default `True`
            Sync the data model
        :param full_export: `bool`, default `True`
            Do a resync of the data model
        :param delete_previous: `bool`, default `False`
            Delete previous exported files.
        :return: None
        """
        _deprecation_msgs("This function was removed from pyCarol")
        return None

    def export_all(self, sync_dm=True, full_export=False, delete_previous=False):
        """

        @DEPRECATED. This function was removed in pycarol 3.34

        Export all data models

        This method will trigger or pause the export of the data in the datamodel to
        CDS

        :param sync_dm: `bool`, default `True`
            Sync the data model
        :param full_export: `bool`, default `True`
            Do a resync of the data model
        :param delete_previous: `bool`, default `False`
            Delete previous exported files.
        :return: None
        """

        _deprecation_msgs("This function was removed from pyCarol")
        return None

    def delete(self, dm_id=None, dm_name=None, entity_space='WORKING'):
        # TODO: Check Possible entity_spaces

        if dm_id is None:
            assert dm_name is not None
            resp = self.get_by_name(dm_name)
            dm_id = resp['mdmId']

        url = f"v1/entities/templates/{dm_id}"
        querystring = {"entitySpace": entity_space}

        return self.carol.call_api(url, method='DELETE', params=querystring)

    def _get_name_type_DMs(self, fields):
        f = {}
        for field in fields:
            if field.get('mdmMappingDataType', None) not in ['NESTED', 'OBJECT']:
                f.update({field['mdmName']: field['mdmMappingDataType']})
            else:
                f[field['mdmName']] = self._get_name_type_DMs(
                    field['mdmFields'])
        return f

    def _get_dm_export_stats(self):
        """

        @DEPRECATED. This function was removed in pycarol 3.34

        Get export status for data models

        :return: `dict`
            dict with the information of which data model is exporting its data.
        """

        _deprecation_msgs("This function was removed from pyCarol")
        return None

    def reprocess_rejected(self, dm_name=None, dm_id=None, delete_records=True):
        """
        Reprocess rejected records.

        Args:
            dm_name: `str` default None
                 Data model name
            dm_id: `str` default None
                 Data model ID
            delete_records: `boll` default True
                True to delete and reprocess, False for copy and reprocess.

        Returns: dict
            Task Definition.
        """

        if dm_name is None and dm_id is None:
            raise ValueError('Either dm_name or dm_id must be set.')

        dm_id = dm_id if dm_id else self.get_by_name(dm_name)['mdmId']

        params = {
            'recordType': 'REJECTED',
            'fuzzy': False,
            'queryDescription': {"a": "REJECTED", "q": "*", "p": "1", "o": "ASC", "r": "[]"},
            'deleteRecords': delete_records
        }
        result = self.carol.call_api(
            f'v1/entities/templates/{dm_id}/reprocess', method='POST', params=params)

        return result

    def reprocess(self, dm_name, copy_or_move='move', record_type='ALL', query_filter=None):
        """
        Reprocess records from a data model

        :param dm_name: `str`
            Data model name
        :param copy_or_move: `str`, default `move`
            Either `move` or `copy` to staging.
        :param record_type:  `str`, default `ALL`
            Type of records to reprocess. `All`, `Golden` or `Rejected`
        :param query_filter:  `Filter.Builder Object`, default `None`
            The Filter instance to reprocess the data on.
        :return: None
        """

        if copy_or_move not in ["copy", "move"]:
            ValueError(
                f'copy_or_move should be "copy" or "move". {copy_or_move} was used')

        if copy_or_move == 'copy':
            copy_or_move = False
        else:
            copy_or_move = True

        mdm_id = self.get_by_name(dm_name)['mdmId']

        url_filter = f"v1/entities/templates/{mdm_id}/reprocess"

        query_params = {"recordType": record_type, "fuzzy": "false",
                        "deleteRecords": copy_or_move}

        result = self.carol.call_api(
            url_filter, data=query_filter, params=query_params, method='POST')
        return result

    def send_data(self, data, dm_name=None, dm_id=None, step_size=500, gzip=False, delete_old_records=False,
                  print_stats=True, max_workers=2, async_send=False):
        """
        :param data: pandas data frame, json.
            Data to be send to Carol
        :param dm_name:  `str`, default `None`
            Data model name
        :param dm_id:  `str`, default `None`
            Data model id
        :param step_size: `int`, default `500`
            Number of records to be sent in each iteration. Max size for each batch is 10MB
        :param print_stats: `bool`, default `True`
            If print the status
        :param gzip: `bool`, default `True`
            If send each batch as a gzip file.
        :param delete_old_records: `bool`, default `False`
            Delete previous records in the data model.
        :param max_workers: `int`, default `2`
            To be used with `async_send=True`. Number of threads to use when sending.
        :param async_send: `bool`, default `False`
            To use async to send the data. This is much faster than a sequential send.
        :return: None
        """

        self.gzip = gzip
        extra_headers = {}
        content_type = 'application/json'
        if self.gzip:
            content_type = None
            extra_headers["Content-Encoding"] = "gzip"
            extra_headers['content-type'] = 'application/json'

        if dm_name:
            dm_id = self.get_by_name(dm_name)['mdmId']
        else:
            assert dm_id
            dm_name = self._get(dm_id, by='id')['mdmName']

        if delete_old_records:
            delete_golden(self.carol, dm_name)

        is_df = False
        if isinstance(data, str):
            data = json.loads(data)
            data_size = len(data)
            _sample_json = data[0]
        elif isinstance(data, list):
            data_size = len(data)
            _sample_json = data[0]
        else:
            import pandas as pd
            if isinstance(data, pd.DataFrame):
                is_df = True
                data_size = data.shape[0]
                _sample_json = data.iloc[0].to_json(date_format='iso')

        if (not isinstance(data, list)) and (not is_df):
            data = [data]
            data_size = len(data)

        url = f"v1/entities/templates/{dm_id}/goldenRecords?async=true"

        if async_send:
            loop = asyncio.get_event_loop()
            future = asyncio.ensure_future(async_helpers.send_data_asynchronous(carol=self.carol,
                                                                                data=data,
                                                                                step_size=step_size,
                                                                                url=url,
                                                                                extra_headers=extra_headers,
                                                                                content_type=content_type,
                                                                                max_workers=max_workers,
                                                                                compress_gzip=self.gzip))
            loop.run_until_complete(future)

        else:
            for data_json, cont in stream_data(data=data,
                                               step_size=step_size,
                                               compress_gzip=self.gzip):

                self.carol.call_api(url, data=data_json, extra_headers=extra_headers,
                                    content_type=content_type, status_forcelist=[
                                        502, 429, 502],
                                    method_whitelist=frozenset(['POST']))
                if print_stats:
                    print('{}/{} sent'.format(cont, data_size), end='\r')

    def create_mapping(self, staging_name, connector_id=None, connector_name=None, dm_name=None, dm_id=None,
                       publish=False):
        """

        This method will create the link between the staging and a data model. If Publish=True it will publish how it is.

        :param staging_name:  `str`
            Name of the staging to be mapped.
        :param connector_id:  `str`, `str`, default `None`
            Staging's connector ID
        :param connector_name:  `str`, default `None`
            Staging's connector name
        :param dm_name: `int`, default `100`
            Data Model name
        :param dm_id: `bool`, default `True`
            Data model ID
        :param publish: `bool`, default `True`
            If publish after create mapping.
        :return: None
        """

        if connector_name:
            _con = Connectors(self.carol)
            connector_id = _con.get_by_name(connector_name)['mdmId']
        else:
            assert connector_id
            _con = Connectors(self.carol)
            connector_name = _con.get_by_id(connector_id)['mdmName']

        if dm_name:
            dm_id = self.get_by_name(dm_name)['mdmId']
        else:
            assert dm_id
            dm_name = self.get_by_id(dm_id)['mdmName']

        url = f"v1/connectors/{connector_id}/entityMappings"

        payload = {"mdmMasterEntityId": dm_id,
                   "mdmMasterEntityName": dm_name,
                   "mdmConnectorId": connector_id,
                   "mdmStagingType": staging_name
                   }
        resp = self.carol.call_api(url, data=payload)
        if publish:
            url = f"v1/connectors/{connector_id}/entityMappings/{resp['mdmId']}/publish"

            self.carol.call_api(url, method='POST')

        return resp

    def fetch_rejected(
            self, dm_name, backend='pandas',
            return_dask_graph=False, callback=None,
            max_hits=None, max_workers=None, file_pattern=None,
            return_callback_result=False
    ):
        """
        Fetch rejected records from Golden.

            Args:
                dm_name: `str`
                    Data model name to be imported
                backend: ['dask','pandas'], default `dask`
                    if to use either dask or pandas to fetch the data
                return_dask_graph: `bool`, default `false`
                    If to return the dask graph or the dataframe.
                callback: `callable`, default `None`
                    Function to be called each downloaded file.
                max_hits: `int`, default `None`
                    Number of records to get.
                max_workers: `int` default `None`
                    Number of workers to use when downloading parquet files with pandas back-end.
                file_pattern: `str` default `None`
                    File pattern to filter data when fetching from CDS. It is possible to fetch only records from a staging table. e.g.
                    file_pattern='{connector_id}_{staging_name}/2019-11-25' will fetch only CDS files that came from
                    the connector  "connector_id" and staging "staging_name" on "2019-11-25".
                return_callback_result `bool` default `False`
                    If a callback is used, it will return the result of the response of the callback. This will skip all the
                    operation to merge records and return selected columns.
                :return:
                    pd.DataFrame with rejected records.
                """

        if backend not in ('dask', 'pandas'):
            raise ValueError(
                f"Backend options are 'dask','pandas' {backend} was given")

        if callback and not callable(callback):
            raise TypeError(f'"{callback}" object is not callable')

        if return_dask_graph and backend != 'dask':
            warnings.warn(
                '`return_dask_graph` has no use when `backend!=dask`')

        import_type = 'golden_rejected'

        storage = Storage(self.carol)
        token_carolina = storage.backend.carolina.token
        storage_space = storage.backend.carolina.get_bucket_name(import_type)

        if backend == 'dask':
            d = _import_dask(
                storage=storage, dm_name=dm_name,
                import_type=import_type,
                merge_records=False,
                return_dask_graph=return_dask_graph,
                columns=None, file_pattern=file_pattern,
            )

        elif backend == 'pandas':
            import pandas as pd
            d = _import_pandas(
                storage=storage, dm_name=dm_name,
                import_type=import_type, columns=None,
                callback=callback, max_hits=max_hits,
                max_workers=max_workers,
                token_carolina=token_carolina,
                storage_space=storage_space, file_pattern=file_pattern
            )
            if d is None:
                warnings.warn("No data to fetch!", UserWarning)
                _field_types = self._get_name_type_DMs(
                    self.get_by_name(dm_name)['mdmFields'])
                cols_keys = list(_field_types)

                d = pd.DataFrame(columns=cols_keys)
                for key, value in _field_types.items():
                    if isinstance(value, dict):
                        value = "STRING"  # If nested we receive as a `STR`
                    d.loc[:, key] = d.loc[:, key].astype(
                        _DATA_MODEL_TYPES_MAPPING.get(value.lower(), str), copy=False)
                return d

        else:
            raise ValueError(
                f'backend should be either "dask" or "pandas" you entered {backend}')

        if (return_callback_result) and (callback is not None):
            return d

        return d

    def generate_SQL_tables(self,):
        """Geneates SQL tables for all data models. 

        The tables created will be named:
        `dm_{datamodel_name}`. 

        Returns:
            dict: Carol's response
        """

        return self.carol.call_api(path='v1/admin/entities/templates/generateSqlTables', method='POST')

    def remove_SQL_tables(self,):
        """Removes SQL tables for all data models. 

        Returns:
            dict: Carol's response
        """

        return self.carol.call_api(path='v1/admin/entities/templates/removeSqlTables', method='POST') 


class entIntType(object):
    ent_type = 'long'


class entDoubleType(object):
    ent_type = 'double'


class entStringType(object):
    ent_type = "string"


class entNullType(object):
    ent_type = "string"


class entBooleanType(object):
    ent_type = "boolean"


class entArrayType(object):
    ent_type = "nested"


class entObjectType(object):
    ent_type = "object"


class entType(object):
    @classmethod
    def get_ent_type_for(cls, t):
        """docstring for get_schema_type_for"""
        SCHEMA_TYPES = {
            type(None): entNullType,
            str: entStringType,
            int: entIntType,
            float: entDoubleType,
            bool: entBooleanType,
            list: entArrayType,
            dict: entObjectType,
        }

        schema_type = SCHEMA_TYPES.get(t)

        if not schema_type:
            raise JsonEntTypeNotFound("There is no schema type for  %s.\n Try:\n %s" % (
                str(t), ",\n".join(["\t%s" % str(k) for k in SCHEMA_TYPES.keys()])))
        return schema_type


class JsonEntTypeNotFound(Exception):
    pass


class CreateDataModel(object):
    def __init__(self, carol):
        self.carol = carol
        self.template_dict = {}

        self.fields = DataModelFields(self.carol)
        self.fields.possible_types()
        self.all_possible_types = self.fields._possible_types
        self.all_possible_fields = self.fields.fields_dict

    def from_snapshot(self, snapshot, publish=False, overwrite=False):

        _count = 0

        while True:
            url = 'v1/entities/templates/snapshot'
            resp = self.carol.call_api(
                path=url, method='POST', data=snapshot, errors='ignore')

            if ('already exists' in resp.get('errorMessage', 'asdf')) and (overwrite):
                del_DM = DataModel(self.carol)
                del_DM.get_by_name(snapshot['entityTemplateName'])
                dm_id = del_DM.entity_template_.get(
                    snapshot['entityTemplateName']).get('mdmId', None)
                if dm_id is None:  # if None
                    continue
                entity_space = del_DM.entity_template_.get(
                    snapshot['entityTemplateName'])['mdmEntitySpace']
                del_DM.delete(dm_id=dm_id, entity_space=entity_space)
                time.sleep(0.5)  # waint for deletion
                _count += 1
                if _count > 5:
                    print(
                        f"Something wrong coping {snapshot['entityTemplateName']}")
                    print(f"Data model was not copied: {resp}")
                    return
                continue

            break
        print('Data Model {} created'.format(snapshot['entityTemplateName']))
        self.template_dict.update({resp['mdmName']: resp})
        if publish:
            self.publish_template(resp['mdmId'])

    def publish_template(self, dm_id):
        url = f'v1/entities/templates/{dm_id}/publish'
        resp = self.carol.call_api(path=url, method='POST')
        return resp

    def _check_verticals(self):
        self.verticals_dict = Verticals(self.carol).all()

        if self.vertical_ids is not None:
            for key, value in self.verticals_dict.items():
                if value == self.vertical_ids:
                    self.vertical_ids = value
                    self.vertical_names = key
                    return
        else:
            for key, value in self.verticals_dict.items():
                if key == self.vertical_names:
                    self.vertical_ids = value
                    self.vertical_names = key
                    return

        raise Exception('{}/{} are not valid values for mdmVerticalNames/mdmVerticalIds./n'
                        ' Possible values are: {}'.format(self.vertical_names, self.vertical_ids,
                                                          self.verticals_dict))

    def _check_entity_template_types(self):
        self.template_type_dict = DataModelTypeIds(self.carol).all()

        if self.entity_template_type_ids is not None:
            for key, value in self.template_type_dict.items():
                if value == self.entity_template_type_ids:
                    self.entity_template_type_ids = value
                    self.entity_template_type_names = key
                    return
        else:
            for key, value in self.template_type_dict.items():
                if key == self.entity_template_type_names:
                    self.entity_template_type_ids = value
                    self.entity_template_type_names = key
                    return

        raise Exception('{}/{} are not valid values for mdmEntityTemplateTypeNames/mdmEntityTemplateTypeIds./n'
                        ' Possible values are: {}'.format(self.vertical_names, self.vertical_ids,
                                                          self.template_type_dict))

    def _check_dm_name(self):

        est_ = DataModel(self.carol)
        try:
            est_.get_by_name(self.dm_name)
        except CarolApiResponseException:
            return

        raise Exception('mdm name {} already exist'.format(self.dm_name))

    def create(self, dm_name, overwrite=False, vertical_ids=None, vertical_names=None, entity_template_type_ids=None,
               entity_template_type_names=None, label=None, group_name='Others',
               transaction_data_model=False):

        self.dm_name = dm_name
        self.group_name = group_name

        if not label:
            self.label = self.dm_name
        else:
            self.label = label
        self.transaction_data_model = transaction_data_model

        assert ((vertical_names is not None) or (vertical_ids is not None))
        assert ((entity_template_type_ids is not None)
                or (entity_template_type_names is not None))

        self.vertical_names = vertical_names
        self.vertical_ids = vertical_ids
        self.entity_template_type_ids = entity_template_type_ids
        self.entity_template_type_names = entity_template_type_names

        self._check_verticals()
        self._check_entity_template_types()
        if not overwrite:
            self._check_dm_name()

        payload = {"mdmName": self.dm_name, "mdmGroupName": self.group_name, "mdmLabel": {"en-US": self.label},
                   "mdmVerticalIds": [self.vertical_ids],
                   "mdmEntityTemplateTypeIds": [self.entity_template_type_ids],
                   "mdmTransactionDataModel": self.transaction_data_model, "mdmProfileTitleFields": []}

        while True:
            url_filter = "v1/entities/templates"
            resp = self.carol.call_api(
                url_filter, data=payload, method='POST', errors='ignore')
            # error handler for token
            if ('already exists' in resp.get('errorMessage', [])) and (overwrite):
                del_DM = DataModel(self.carol)
                del_DM.get_by_name(self.dm_name)
                dm_id = del_DM.entity_template_.get(
                    self.dm_name).get('mdmId', None)
                if dm_id is None:  # if None
                    continue
                entity_space = del_DM.entity_template_.get(self.dm_name)[
                    'mdmEntitySpace']
                del_DM.delete(dm_id=dm_id, entity_space=entity_space)
                time.sleep(0.5)  # waint for deletion
                continue
            break

        self.template_dict.update({resp['mdmName']: resp})

    def _profile_title(self, profile_title, dm_id):
        if isinstance(profile_title, str):
            profile_title = [profile_title]

        profile_title = [i.lower() for i in profile_title]

        url = f"v1/entities/templates/{dm_id}/profileTitle"
        resp = self.carol.call_api(path=url, method='POST', data=profile_title)
        return resp

    def add_field(self, field_name, dm_id=None, parent_field_id=""):

        if dm_id is None:
            assert self.dm_id
        else:
            est_ = DataModel(self.carol)
            est_.get_by_id(dm_id)
            if est_.entity_template_ == {}:
                print('Template does not exisit')
                return
            self.dm_id = dm_id
            _, template_ = est_.entity_template_.popitem()
            self.current_fields = [
                i for i in template_['mdmFieldsFull'].keys()]
            if field_name.lower() in self.current_fields:
                print("'{}' already in the template".format(field_name))
                return

        field_to_send = self.all_possible_fields.get(field_name)
        if field_to_send is None:
            print('Field does not exist')
            return
        querystring = {"parentFieldId": parent_field_id}

        url = f"v1/entities/templates/{self.dm_id}/onboardField/{field_to_send['mdmId']}"
        resp = self.carol.call_api(path=url, method='POST', params=querystring)

    def _labels_and_desc(self, prop):

        if self.label_map is None:
            label = prop
        else:
            label = self.label_map.get(prop)
            if label is None:
                label = prop
            else:
                label = {"en-US": label}

        if self.description_map is None:
            description = prop
        else:
            description = self.description_map.get(prop)
            if description is None:
                description = prop
            else:
                description = {"en-US": description}

        return label, description

    def from_json(self, json_sample, profile_title=None, publish=False, dm_id=None,
                  label_map=None, description_map=None, ignore_field_type=False):

        if publish:
            assert profile_title is not None, "To publish the data model, `profile_title` has to be set."
            if isinstance(profile_title, str):
                profile_title = [profile_title]
            assert all([i in json_sample for i in profile_title]
                       ), "all profile title values should be in `json_sample`"

        self.label_map = label_map
        self.description_map = description_map

        if dm_id is None:
            assert self.template_dict is not None
            template_name, template_json = self.template_dict.copy().popitem()
            self.dm_id = template_json['mdmId']

        else:
            self.dm_id = dm_id

        self.json_sample = json_sample

        n_fields = len(list(self.json_sample))
        count = 0
        for prop, value in self.json_sample.items():
            count += 1
            print('Creating {}/{}'.format(count, n_fields))
            prop = prop.lower()
            entity_type = entType.get_ent_type_for(type(value))
            if prop in self.all_possible_fields.keys():
                if not entity_type.ent_type == 'nested':
                    ent_ = self.all_possible_fields.get(prop, []).copy()
                    ent_.pop('mdmCreated')
                    ent_.pop('mdmLastUpdated')
                    ent_.pop('mdmTenantId')
                    if (ent_['mdmMappingDataType'].lower() == entity_type.ent_type) or (ignore_field_type):
                        self.add_field(prop, parent_field_id="")
                    else:
                        print('problem, {} not created, field name matches with an already'
                              'created field but different type'.format(prop))
                else:
                    print('Nested fields are not supported')
            else:
                if not entity_type.ent_type == 'nested':

                    current_label, current_description = self._labels_and_desc(
                        prop)
                    self.fields.create(mdm_name=prop, mdm_mpping_data_type=entity_type.ent_type,
                                       mdm_field_type='PRIMITIVE', admin=True,
                                       mdm_label=current_label, mdm_description=current_description)
                    self.all_possible_fields = self.fields.fields_dict
                    self.add_field(prop, parent_field_id="")
                else:
                    print('Nested fields are not supported')

        if publish:
            self._profile_title(profile_title, self.dm_id)
            self.publish_template(self.dm_id)



import copy
from string import Formatter
from datetime import datetime, timedelta
import threading

class Carolina:
    """
    This class is used to handle any iteration with Carol data storage (CDS).

    Args:

        carol: class: pycarol.Carol
    """

    token = None
    backend_lock = threading.Lock()

    def __init__(self, carol):

        self.carol = carol
        self.engine = None
        self.client = None
        self.token = None
        self.expires_at = None
        self.cds_app_storage_path = None
        self.cds_golden_path = None
        self.cds_staging_path = None
        self.cds_staging_master_path = None
        self.cds_staging_rejected_path = None
        self.cds_view_path = None
        self.cds_golden_intake_path = None
        self.cds_staging_intake_path = None
        self.cds_golden_rejected_intake_path = None
        self.cds_view_intake_path = None
        self.cds_pycarol_storage_path = None

    def init_if_needed(self):
        """
        Checks if backend credentials / object is still valid. This method is thread safe by using lock.

        Returns:
            True if backend has been reinitialized, false if it is still valid.

        """
        expired = False

        # locks all instance variables to prevent inconsistencies
        with Carolina.backend_lock:
            if self.client:
                # Check if the token is not expired for at least another minute.. we do a little margin to avoid time difference issues
                if self.expires_at is None or datetime.utcnow() + timedelta(minutes=1) < self.expires_at:
                    return 
                else:
                    expired = True

            if Carolina.token is None or expired or Carolina.token.get('tenant_name', '') != self.carol.tenant['mdmName'] \
                    or Carolina.token.get('app_name', '') != self.carol.app_name \
                    or (datetime.utcnow() + timedelta(minutes=1)
                        >= datetime.fromtimestamp(Carolina.token.get('expirationTimestamp', 1)/1000.0)):
                token = self.carol.call_api('v1/storage/storage/token', params={'carolAppName': self.carol.app_name})
                token['tenant_name'] = self.carol.tenant['mdmName']
                token['app_name'] = self.carol.app_name
                Carolina.token = token

            token = Carolina.token
            self.engine = token['engine']

            if token.get('expirationTimestamp', None) is not None:
                self.expires_at = datetime.fromtimestamp(token['expirationTimestamp'] / 1000.0)

            self.cds_app_storage_path = token['cdsAppStoragePath']
            self.cds_golden_path = token['cdsGoldenPath']
            self.cds_staging_path = token['cdsStagingPath']
            self.cds_staging_master_path = token['cdsStagingMasterPath']
            self.cds_staging_rejected_path = token['cdsStagingRejectedPath']
            self.cds_view_path = token['cdsViewPath']

            self.cds_golden_intake_path = token['cdsIntakeGoldenPath']
            self.cds_staging_intake_path = token['cdsIntakeStagingPath']

            self.cds_golden_rejected_intake_path = token['cdsRejectedPath']

            self.cds_pycarol_storage_path = copy.deepcopy(token['cdsAppStoragePath'])
            self.cds_pycarol_storage_path['path'] = 'carolapps/._pycarol' 

            self.cds_view_intake_path = token['cdsIntakeViewPath']

            if self.engine == 'GCP-CS':
                self._init_gcp(token)

    def _init_gcp(self, token):
        """
        Initialize GCP back-end.
        Warning: this method is not thread safe and should not be called directly, use init_if_needed instead.

        Args:
            token: `dict`
            TODO: add here the fields for GCP
                GCP token
                {
                    "aiAccessKeyId": aws_access_key_id,
                    "aiSecretKey": aws_secret_access_key,
                    "aiAccessToken": aws_session_token,
                }

        Returns:
            None

        """
        from google.oauth2 import service_account
        from google.cloud import storage

        self.token = token['token']
        gcp_credentials = service_account.Credentials.from_service_account_info(token['token'])
    
        self.client = storage.Client(credentials=gcp_credentials, project=token['token']['project_id'])

    def get_client(self):
        self.init_if_needed()
        return self.client

    def get_bucket_name(self, space):
        """
        Format the bucket path for each possible space.

        Args:

            space:  `str`,
                Which bucket to get. Possible values:

                    1. "golden": Data Model golden records.
                    2. "staging": Staging records path
                    3. "staging_master": Staging records from Master
                    4. "staging_rejected": Staging records from Rejected
                    5. "view": Data Model View records
                    6. "app": App  bucket
                    7. "golden_cds": CDS golden records
                    8. "staging_cds": Staging Intake.

        Returns:
            formatted bucket path

        """

        # TODO: we can use a dictionary or instead of ifs.
        if space == 'golden':
            template = self.cds_golden_path['bucket']
        elif space == 'staging':
            template = self.cds_staging_path['bucket']
        elif space == 'staging_master':
            template = self.cds_staging_master_path['bucket']
        elif space == 'staging_rejected':
            template = self.cds_staging_rejected_path['bucket']
        elif space == 'view':
            template = self.cds_view_path['bucket']
        elif space == 'app':
            template = self.cds_app_storage_path['bucket']
        elif space == 'staging_cds':
            template = self.cds_staging_intake_path['bucket']
        elif space == 'golden_cds':
            template = self.cds_golden_intake_path['bucket']
        elif space == 'view_cds':
            template = self.cds_view_intake_path['bucket']
        elif space == 'golden_rejected':
            template = self.cds_staging_rejected_path['bucket']
        elif space == 'pycarol':
            template = self.cds_pycarol_storage_path['bucket']
        else:
            raise ValueError

        name = Formatter().vformat(template, None, {'tenant_id': self.carol.tenant['mdmId']})
        return name

    def get_path(self, space, vars):
        """
        Format the bucket path for each possible space.

        Args:

            space:  `str`,
                Which bucket to get. Possible values:

                1. "golden": Data Model golden records.
                2. "staging": Staging records path
                3. "staging_master": Staging records from Master
                4. "staging_rejected": Staging records from Rejected
                5. "view": Data Model Relationship View records
                6. "app": App  bucket
                7. "golden_cds": CDS golden records
                8. "staging_cds": Staging Intake.
            vars: `dict`
                Parameters needed to format the storage path. Possible keys:

                1. "tenant_id": Tenant ID.
                2. "connector_id": Connector ID
                3. "staging_type": Staging Name
                4. "dm_name": Data model Name
                5. "relationship_view_name": Relationship view name
                6. "app_name": App  name


        Returns:
            formatted path

        """
        vars['tenant_id'] = self.carol.tenant['mdmId']

        # TODO: we can use a dictionary or instead of ifs.
        if space == 'golden':
            template = self.cds_golden_path['path'] + '/'
        elif space == 'staging':
            template = self.cds_staging_path['path'] + '/'
        elif space == 'staging_master':
            template = self.cds_staging_master_path['path'] + '/'
        elif space == 'staging_rejected':
            template = self.cds_staging_rejected_path['path'] + '/'
        elif space == 'view':
            template = self.cds_view_path['path'] + '/'
        elif space == 'app':
            template = self.cds_app_storage_path['path'] + '/'
        elif space == 'staging_cds':
            template = self.cds_staging_intake_path['path'] + '/'
        elif space == 'golden_cds':
            template = self.cds_golden_intake_path['path'] + '/'
        elif space == 'view_cds':
            template = self.cds_view_intake_path['path'] + '/'
        elif space == 'golden_rejected':
            template = self.cds_golden_rejected_intake_path['path'] + '/'
        elif space == 'pycarol':
            template = self.cds_pycarol_storage_path['path'] + '/'
        else:
            raise ValueError

        name = Formatter().vformat(template, None, vars)
        return name




import warnings
from urllib3.util.retry import Retry
import requests
from requests.adapters import HTTPAdapter
import json
import os, copy
import os.path
from .auth.ApiKeyAuth import ApiKeyAuth
from .auth.PwdAuth import PwdAuth
from .tenant import Tenant
from . import __CONNECTOR_PYCAROL__
from . import __version__
from . organization import Organization
from .exceptions import CarolApiResponseException, InvalidToken

class Carol:
    """
    Handles all Carol`s API calls It will handle all API calls,
    for a given authentication method. :param domain: `str`.

    Args:

        domain: `str`. default `None`.
            Tenant name. e.x., domain.carol.ai
        app_name: `str`. default `None`.
            Carol app name.
        auth: `PwdAuth` or `ApiKeyAuth`.
            object Auth Carol object to handle authentication
        connector_id: `str` , default `__CONNECTOR_PYCAROL__`.
            Connector Id
        port: `int` , default 443.
            Port to be used (when running locally it could change)
        verbose: `bool` , default `False`.
            If True will print the header, method and URL of each API call.
        organization: `str` , default `None`.
            Organization domain.
        environment: `str`, default `carol.ai`,
            Which Carol's environment to use. There are three possible values today.

                1. 'carol.ai' for the production environment
                2. 'karol.ai' for the explore environment
                3. 'qarol.ai' for the QA environment

        host: `str` default `None`
            This will overwrite the host used. Today the host is:

                1. if organization is None, host={domain}.{environment}
                2. else host={organization}.{environment}

            See Carol._set_host.
        
         user:  `str` default `None`
            User
         password: `str` default `None`
            User passowrd
         api_key:  `str` default `None`
            Carol's Api Key
        org_level: `bool` default `False`
            If True, will log-in at organization level.

    OBS:
        In case all parameters are `None`, pycarol will try yo find their values in the environment variables.
        The values are:

             1. `CAROLTENANT` for domain
             2. `CAROLAPPNAME` for app_name
             3. `CAROLAPPOAUTH` for auth
             4. `CAROLORGANIZATION` for organization
             5. `CAROLCONNECTORID` for connector_id
             6. `CAROL_DOMAIN` for environment
             7. `CAROLUSER` for carol user email
             8. `CAROLPWD` for user password.


    """

    def __init__(self, domain=None, app_name=None, auth=None, connector_id=None, port=443, verbose=False,
                 organization=None, environment=None, host=None, user=None, password=None, api_key=None, org_level=False):

        self.connector_id = connector_id

        if auth is None:
            auth = self.__get_auth(connector_id=connector_id, username=user, password=password, app_oauth=api_key)

        if (domain is None) and (not org_level):
            domain = os.getenv('CAROLTENANT')
            if domain is None:
                raise ValueError("`domain` must be set.")

        if org_level:
            domain = None

        if app_name is None:
            app_name = os.getenv('CAROLAPPNAME', ' ')

        if self.connector_id is None:
            if auth.connector_id is None:
                self.connector_id = __CONNECTOR_PYCAROL__
            else:
                self.connector_id = auth.connector_id

        if (domain is None and not org_level) or app_name is None or auth is None:
            raise ValueError("domain, app_name and auth must be specified as parameters, either " +
                             "in the environment variables CAROLTENANT, CAROLAPPNAME, CAROLAPPOAUTH" +
                             " or CAROLUSER+CAROLPWD and  CAROLCONNECTORID")

        # TODO Fixed to be compatible with the old `ENV_DOMAIN`. We could add a deprecated warning.
        self.environment = environment if environment is not None else os.getenv('CAROL_DOMAIN',
                                                                                 os.getenv('ENV_DOMAIN', 'carol.ai'))
        self.organization = organization if organization is not None else os.getenv('CAROLORGANIZATION',
                                                                                    os.getenv('API_SUBDOMAIN'))
        self.domain = domain
        self.app_name = app_name
        self.port = port
        self.verbose = verbose
        self._host_string = host
        self.host = self._set_host(domain=self.domain, organization=self.organization,
                                   environment=self.environment, host=host)
        self._tenant = None
        self._current_user = None
        self.auth = auth
        self.auth.set_connector_id(self.connector_id)
        self.session = None
        self._is_org_level = org_level

        self.auth.login(self)
        self.response = None

        self.org = None

    def __get_auth(self, connector_id=None, username=None, password=None, app_oauth=None):

        if username is None:
            username = os.getenv('CAROLUSER')

        if password is None:
            password = os.getenv('CAROLPWD')

        if username and password:
            return PwdAuth(user=username, password=password)
            
        if app_oauth is None:
            app_oauth = os.getenv('CAROLAPPOAUTH')
        
        if connector_id is None:
            connector_id = os.getenv('CAROLCONNECTORID', __CONNECTOR_PYCAROL__)
            self.connector_id = connector_id
        
        if app_oauth and connector_id:
            return ApiKeyAuth(app_oauth)

        raise ValueError("either `auth` or `username/password` or `api_key` or pycarol env variables must be set.")

    @property
    def current_user(self):
        """
        Returns the current user.
        
        Args:
            force: `bool` default `False`.
                If True will force the request to fetch current user.
        """

        if self._current_user is None:
            if self._is_org_level:
                self._current_user = self.call_api('v2/orgUsers/current', )
            else:
                self._current_user = self.call_api('v2/users/current', )
        return self._current_user


    @property
    def tenant(self):
        if self._tenant is None and not self._is_org_level:
            self._tenant = Tenant(self).get_tenant_by_domain(self.domain)
        return self._tenant


    @staticmethod
    def _set_host(domain, organization, environment, host):
        """
        Set the host to be used.

        Args:

            domain: `str`
                Former tenant name.
                e.x., domain.carol.ai
            organization: `str`
                Organization domain.
            environment: `str`
                Which Carol's environment to use. There are three possible values today.

                1. 'carol.ai' for the production environment
                2. 'karol.ai' for the explore environment
                3. 'qarol.ai' for the QA environment

            host: `str`
                Host to be used. It overwrite the default one.

        Returns: `str`
            host
        """
        if host is not None:
            return host
        elif organization is not None:
            return f"{organization}.{environment}"
        else:
            return f"{domain}.{environment}"

    @staticmethod
    def _retry_session(retries=5, session=None, backoff_factor=0.5, status_forcelist=(500, 502, 503, 504, 524),
                       method_whitelist=frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE'])):

        """
        Static method used to handle retries between calls.

        Args:

            retries: `int` , default `5`
                Number of retries for the API calls
            session: Session object dealt `None`
                It allows you to persist certain parameters across requests.
            backoff_factor: `float` , default `0.5`
                Backoff factor to apply between  attempts. It will sleep for:
                        {backoff factor} * (2 ^ ({retries} - 1)) seconds
            status_forcelist: `iterable` , default (500, 502, 503, 504, 524).
                A set of integer HTTP status codes that we should force a retry on.
                A retry is initiated if the request method is in method_whitelist and the response status code is in
                status_forcelist.
            method_whitelist: `iterable` , default frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE']))
                Set of uppercased HTTP method verbs that we should retry on.

        Returns:
            :class:`requests.Section`
        """

        session = session or requests.Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
            method_whitelist=method_whitelist,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        return session

    def call_api(self, path, method=None, data=None, auth=True, params=None, content_type='application/json', retries=8,
                 session=None, backoff_factor=0.5, status_forcelist=(502, 503, 504, 524), downloadable=False,
                 method_whitelist=frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE', 'POST']), errors='raise',
                 extra_headers=None, files=None, prefix_path='/api/',
                 **kwds):
        """
        This method handles all the API calls.

        Args:

            path: `str`.
                API URI path. e.x.  v2/staging/schema
            method: 'str', default `None`.
                Set of uppercased HTTP method verbs that we should call on.
            data: 'dict`, default `None`.
                Dictionary, list of tuples, bytes, or file-like object to send in
                the body of the request.
            auth: :class: `pycarol.ApiKeyAuth` or `pycarol.PwdAuth`
                Auth type to be used within the API's calls.
            params: (optional) Dictionary, list of tuples or bytes to send
                     in the query string for the :class:`requests.Request`.
            content_type: `str`, default 'application/json'
                Content type for the api call
            retries: `int` , default `5`
                Number of retries for the API calls
            session: :class `requests.Session` object dealt `None`
                It allows you to persist certain parameters across requests.
            backoff_factor: `float` , default `0.5`
                Backoff factor to apply between  attempts. It will sleep for:
                        {backoff factor} * (2 ^ ({retries} - 1)) seconds
            status_forcelist: `iterable` , default (500, 502, 503, 504, 524).
                A set of integer HTTP status codes that we should force a retry on.
                A retry is initiated if the request method is in method_whitelist and the response status code is in
                status_forcelist.
            downloadable: `bool` default `False`.
                If the request will return a file to download.
            method_whitelist: `iterable` , default frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE']))
                Set of uppercased HTTP method verbs that we should retry on.
            errors: {‘ignore’, ‘raise’}, default ‘raise’
                If ‘raise’, then invalid request will raise an exception If ‘ignore’,
                then invalid request will return the request response
            extra_headers: `dict` default `None`
                extra headers to be sent.
            files: `dict` default `None`
                Used when uploading files to carol. This will be sent to :class: `requests.request`
            prefix_path: `str` default `/api/`
                Prefix path to be used to create the final url 'https://{self.host}:{self.port}{prefix_path}{path}.
            kwds: `dict` default `None`
                Extra parameters to be sent to :class: `requests.request`

        Rerturn:
            Dict with API response.

        """
        if session is not None:
            self.session = session

        extra_headers = extra_headers or {}
        url = f'https://{self.host}:{self.port}{prefix_path}{path}'

        if method is None:
            if data is None:
                method = 'GET'
            else:
                method = 'POST'

        met_list = ['HEAD', 'TRACE', 'GET', 'PUT','POST', 'OPTIONS', 'PATCH',
                    'DELETE', 'CONNECT' ]
        assert method in met_list, f"API method must be {met_list}"

        headers = {'accept': 'application/json'}
        if auth:
            self.auth.authenticate_request(headers)

        data_json = None
        if method == 'GET':
            pass

        elif (method == 'POST') or (method == 'DELETE') or (method == 'PUT'):
            if content_type is not None:
                headers['content-type'] = content_type

            if content_type == 'application/json':
                data_json = data
                data = None

        headers.update(extra_headers)
        headers.update({'User-Agent': f'pyCarol/{__version__}'})

        __count = 0
        while True:
            self.session = self._retry_session(
                retries=retries, session=self.session, backoff_factor=backoff_factor,
                status_forcelist=status_forcelist, method_whitelist=method_whitelist
                )

            response = self.session.request(method=method, url=url, data=data, json=data_json,
                                       headers=headers, params=params, files=files, **kwds)

            if self.verbose:
                if data_json is not None:
                    print("Calling {} {}. Payload: {}. Params: {}".format(method, url, data_json, params))
                else:
                    print("Calling {} {}. Payload: {}. Params: {}".format(method, url, data, params))
                print("        Headers: {}".format(headers))

            if response.ok or errors == 'ignore':
                if downloadable:  #Used when downloading carol app file.
                    return response

                response.encoding = 'utf-8'
                self.response = response
                if response.text == '':
                    return {}
                return json.loads(response.text)

            elif (response.reason == 'Unauthorized') and isinstance(self.auth, PwdAuth):
                if response.json().get('possibleResponsibleField') in ['password', 'userLogin']:
                    raise InvalidToken(response.text)
                self.auth.get_access_token()  #It will refresh token if Unauthorized
                __count += 1
                if __count < 5: #To avoid infinity loops
                    continue
                else:
                    raise Exception('Too many retries to refresh token.\n', response.text, response.status_code)
            elif response.status_code == 404:
                raise CarolApiResponseException(response.text, response.status_code)

            raise Exception(response.text, response.status_code)

    def issue_api_key(self, connector_id=None):
        """
        Create an API key for a given connector.

        Args:

            connector_id: `str` default `None`.
                Connector ID to be used when creating the APIkey

        Returns: `dict`
            Dictionary with the API key.

        """

        if connector_id is None:
            connector_id = self.connector_id
        resp = self.call_api('v2/apiKey/issue', data={
            'connectorId': connector_id
        }, content_type='application/x-www-form-urlencoded')
        return resp

    def api_key_details(self, api_key, connector_id):

        """
        Display information about the API key.

        Args:

            api_key: `str`
                Carol's api key
            connector_id: `str`
                Connector Id which API key was created.

        Returns: `dict`
            Dictionary with API key information.

        """

        resp = self.call_api('v2/apiKey/details',
                             params = {"apiKey": api_key, "connectorId": connector_id})

        return resp

    def api_key_revoke(self, connector_id):
        """
        Revoke API key for ta given connector_id

        Args:

            connector_id: `str`
                Connector Id which API key was created.

        Returns: `dict`

            Dictionary with API request response.

        """

        resp = self.call_api('v2/apiKey/revoke', method='DELETE',
                             content_type='application/x-www-form-urlencoded',
                             params = {"connectorId": connector_id})

        return resp

    def switch_org_level(self):
        """
        Switch organization level.

        """
        if self._is_org_level:
            warnings.warn(
                "already in org level.",
                UserWarning, stacklevel=3
            )
            return
        org = self._current_org()
        self.auth.switch_org_context(org['mdmId'])
        self._is_org_level = True

    def switch_environment(self, env_name=None, env_id=None, app_name=None, org_name=None, org_id=None):
        """
        Switch org/environments. If the user has access to this environment, it will be "logged in" in this new
        org/environment.

        Args:

            env_name: `str` default `None`
                Environment (tenant) name to switch the context to.
            env_id: `str` default `None`
                Environment (tenant) id to switch the context to.
            app_name: `str` default `None`
                App name in the target environment to switch the context to.
                Only needed with using CDS.
            org_name: `str` default `None`
                The organization name to switch context to. If the same keep it `None`
            org_id: `str` default `None`
                The organization id to switch context to. If the same keep it `None`

        Returns:
            self

        .. code:: python

            from pycarol import Carol, Staging
            carol = Carol('B', 'teste', auth=PwdAuth('email@totvs.com.br', 'pass'), )
            carol.switch_environment('A')
            Staging(carol_tenant_A).fetch_parquet(...) # fetch parquet from tenant A
            # To switch back
            carol.switch_environment('B')
            #back to tenant B


        """

        if self.org is None:
            self.org = Organization(self).get_organization_info(self.organization)

        current = self.get_current()

        if (org_id is not None and org_id != current['org_id']) or (org_name is not None and org_name != current['org_name'] ):
            # Switch to org context.
            if not current['org_level']:
                self.switch_org_level()
            if org_id is None:
                org_id = Organization(self).get_organization_info(org_name)['mdmId']
            # Switch org.
            self.auth.switch_org_context(org_id)

        if env_name:
            env_id = Tenant(self).get_tenant_by_domain(env_name)['mdmId']
        if env_id is not None:
            self.auth.switch_context(env_id=env_id)
            self._tenant = self._current_env()
            self._is_org_level = False
        else:
            self._tenant = None
            self._is_org_level = True

        self.domain = env_name
        self.app_name = app_name # TODO: Today we cannot use CDS without a valid app name.
        self.organization = self._current_org()['mdmName']

        self.host = self._set_host(domain=self.domain, organization=self.organization,
                                   environment=self.environment, host=self._host_string)

        return self

    def switch_context(self, env_name=None, env_id=None, app_name=None, org_name=None, org_id=None):
        """
        Context manager to temporary have access to a second environment

        Args:

            env_name: `str` default `None`
                Environment (tenant) name to switch the context to.
            env_id: `str` default `None`
                Environment (tenant) id to switch the context to.
            app_name: `str` default `None`
                App name in the target environment to switch the context to.
                Only needed with using CDS.

        Returns:
            None

        Usage:

        .. code:: python

            from pycarol import Carol, Staging
            carol = Carol('B', 'teste', auth=PwdAuth('email@totvs.com.br', 'pass'), )
            with carol.switch_context('A') as carol_tenant_A:
                Staging(carol_tenant_A).fetch_parquet(...) # fetch parquet from tenant A
            #back to tenant B


        """

        if self.org is None:
            self.org = Organization(self).get_organization_info(self.organization)

        if env_name:
            env_id = Tenant(self).get_tenant_by_domain(env_name)['mdmId']

        class SwitchContext(object):

            def __init__(self, parent_context, env_name=None, env_id=None, org_id=None, org_name=None, app_name=None):
                self.parent_context = parent_context
                self.env_name = env_name
                self.env_id = env_id
                self.app_name = app_name
                self.org_id = org_id
                self.org_name = org_name

            def __enter__(self):
                self.parent_context.switch_environment(env_name=self.env_name, env_id=self.env_id,
                                                       app_name=self.app_name, org_name=self.org_name,
                                                       org_id=self.org_id)
                return self.parent_context

            def __exit__(self, exc_type, exc_val, exc_tb):
                del self.parent_context

        return SwitchContext(parent_context=copy.deepcopy(self), env_name=env_name, env_id=env_id, app_name=app_name,
                             org_name=org_name, org_id=org_id,)


    def _current_env(self):
        return self.call_api('v1/tenants/current', errors='ignore')

    def _current_org(self):
        return self.call_api('v1/organizations/current')

    def get_current(self, level='all'):
        """
        Get current org/env information.

        Args:

            level: `str`:
                Possible Values:
                    "all": To get organization and environment information.
                    "org": To get organization information.
                    "env": To get environment information.

        Returns: `dict`
            Dictionary with keys org_Id, org_name, env_id, env_name

        """

        env = {}
        org = {}

        if level.lower() not in ['org', 'env', 'all']:
            raise ValueError(f"level should be 'all', 'org', 'env', {level} was passed.")

        if level.lower() == 'env' or level.lower() == 'all':
            env = self._current_env()

        if level.lower() == 'org' or level.lower() == 'all':
            org = self._current_org()

        if env.get('mdmName') is None:
            self._is_org_level = True
        else:
            self._is_org_level = False

        return {
            "env_name": env.get('mdmName'),
            "env_id": env.get('mdmId'),
            "org_name": org.get('mdmName'),
            "org_id": org.get('mdmId'),
            "org_level": self._is_org_level
        }

    def get_tenants_for_user(self):
        """
        Get all tenants for the current user.

        Returns: `list`
            List of tenants.

        """

        return self.call_api("v1/users/assignedTenantsForCurrentUser", method='GET')



import json


class DataModelTypeIds(object):
    def __init__(self, carol):

        self.carol = carol
        self.offset = 0
        self.page_size = 100
        self.sort_order = 'ASC'
        self.sort_by = None

    def _build_query_params(self):
        if self.sort_by is None:
            self.query_params = {"offset": self.offset, "pageSize": str(self.page_size), "sortOrder": self.sort_order}
        else:
            self.query_params = {"offset": self.offset, "pageSize": str(self.page_size), "sortOrder": self.sort_order,
                                "sortBy": self.sort_by}


    def all(self, offset=0, page_size=100, sort_order='ASC', sort_by=None, print_status=True, save_file=False,
               filename='data/DataModelTypeIds.json'):

        self.offset = offset
        self.page_size = page_size
        self.sort_order = sort_order
        self.sort_by = sort_by
        self._build_query_params()

        self.template_type_dict = {}
        self.template_type_data = []
        count = self.offset


        set_param = True
        self.total_hits = float("inf")
        if save_file:
            file = open(filename, 'w', encoding='utf8')
        while count < self.total_hits:
            url_filter = "v1/entityTemplateTypes"
            query = self.carol.call_api(url_filter, params=self.query_params,)

            count += query['count']
            if set_param:
                self.total_hits = query["totalHits"]
                set_param = False

            query = query['hits']
            self.template_type_data.extend(query)
            self.template_type_dict.update({i['mdmName']: i['mdmId'] for i in query})
            self.query_params['offset'] = count
            if print_status:
                print('{}/{}'.format(count, self.total_hits), end ='\r')
            if save_file:
                file.write(json.dumps(query, ensure_ascii=False))
                file.write('\n')
                file.flush()
        if save_file:
            file.close()
        return self.template_type_dict




class ManifestValidator():

    def __init__(self, manifest_json):
        self.manifest_json = manifest_json


    def run(self):
        status, message = self._basic_processes('batch')
        if status is False:
            return message

        status, message = self._basic_processes('online')
        if status is False:
            return message

        return ''


    def _basic_processes(self, process_type):
        if process_type in self.manifest_json:
            for process in self.manifest_json[process_type]['processes']:
                if 'name' not in process:
                    return False, '"name" not found in some "{}" process'.format(process_type)
                if 'algorithmName' not in process:
                    return False, '"algorithmName" not found in the "{}" process called "{}"'.format(process_type, process['name'])
                if 'algorithmTitle' not in process:
                    return False, '"algorithmTitle" not found in the "{}" process called "{}"'.format(process_type, process['name'])
                if process_type == 'online' and 'properties' not in process['instanceProperties']:
                    return False, '"properties" not found in the "{}" process called "{}"'.format(process_type, process['name'])

                if 'git' in process:
                    for git in process['git']:
                        if "type" not in git:
                            return False, '"type" not found in the "git". "{}" process called "{}"'.format(process_type, process['name'])
                        if "url" not in git:
                            return False, '"url" not found in the "git". "{}" process called "{}"'.format(process_type, process['name'])
                        if "destinationFolder" not in git:
                            return False, '"destinationFolder" not found in the "git". "{}" process called "{}"'.format(process_type, process['name'])
                        if "branch" in git and git['branch'] == '':
                            return False, '"branch" is empty in the "git". "{}" process called "{}"'.format(process_type, process['name'])
        return True, ''



import logging
from collections import defaultdict
import random
import time
from itertools import chain

from pycarol import (
    Carol, ApiKeyAuth, PwdAuth, Tasks, Staging, Connectors, CDSStaging, Subscription, DataModel, Apps, CDSGolden
)
from pycarol.query import delete_golden

def track_tasks(carol, task_list, retry_count=3, logger=None, callback=None, polling_delay=5):
    """Track a list of taks from carol, waiting for errors/completeness. 

    Args:

        carol (pycarol.Carol): pycarol.Carol instance
        task_list (list): List of tasks in Carol
        retry_count (int, optional): Number of times to restart a failed task. Defaults to 3.
        logger (logging.logger, optional): logger to log information. Defaults to None.
        callback (calable, optional): This function will be called every time task status are fetched from carol. 
            A dictionary with task status will be passed to the function. Defaults to None.
        polling_delay (int, optional): Time in seconds to pull task status from Carol

    Usage:

    .. code:: python

        from pycarol import Carol
        from pycarol.functions import track_tasks
        carol = Carol()
        def callback(task_list):
            print(task_list)
        track_tasks(carol=carol, task_list=['task_id_1', 'task_id_2'], callback=callback)

    Returns:

        [dict, bool]: dict with status of each task and booling if any task failed more than retry_count times.
    
    """
    
    if logger is None:
        logger = logging.getLogger(carol.domain)

    retry_tasks = defaultdict(int)
    n_task = len(task_list)
    max_retries = set()
    carol_task = Tasks(carol)
    while True:
        task_status = defaultdict(list)
        for task in task_list:
            status = carol_task.get_task(task).task_status
            task_status[status].append(task)
        for task in task_status['FAILED'] + task_status['CANCELED']:
            logger.warning(f'Something went wrong while processing: {task}')
            retry_tasks[task] += 1
            if retry_tasks[task] > retry_count:
                max_retries.update([task])
                logger.error(
                    f'Task: {task} failed {retry_count} times. will not restart')
                continue

            logger.info(f'Retry task: {task}')
            carol_task.reprocess(task)

        if len(task_status['COMPLETED']) == n_task:
            logger.debug(f'All task finished')
            return task_status, False

        elif len(max_retries) + len(task_status['COMPLETED']) == n_task:
            logger.warning(f'There are {len(max_retries)} failed tasks.')
            return task_status, True
        else:
            time.sleep(round(polling_delay + random.random(), 2))
            logger.debug('Waiting for tasks')
        if callable(callback):
            callback(task_status)


def pause_dm_mappings(carol, dm_list, connector_name=None, connector_id=None, do_not_pause_staging_list=None):
    """Pause mappings from a connetor based on a list of Datamodels.

    Args:

        carol (pycal.Carol): Carol instance
        dm_list (list): list of Datamodels to pause
        connector_name (str): connector name
        connector_name (str): connector id
        do_not_pause_staging_list (list, optional): List of stagings to do not pause. Defaults to None.

    Usage:

    .. code:: python

        from pycarol import Carol
        from pycarol.functions import pause_dm_mappings
        carol = Carol()
        pause_dm_mappings(carol, connector_name='connector_name', dm_list=['dm1','dm2'])

    """

    do_not_pause_staging_list = do_not_pause_staging_list if do_not_pause_staging_list else ['']
    conn = Connectors(carol)
    mappings = conn.get_dm_mappings(connector_name=connector_name, connector_id=connector_id )
    mappings = mappings = [i['mdmId'] for i in mappings if
                           (i['mdmRunningState'] == 'RUNNING') and
                           (i['mdmMasterEntityName'] in dm_list) and
                           (i['mdmStagingType'] not in do_not_pause_staging_list)
                           ]

    _ = conn.pause_mapping(
        connector_name=connector_name, entity_mapping_id=mappings, 
        connector_id=connector_id)

def check_mapping(login, staging_name, connector_name=None, connector_id=None):
    """Check if a staging is mapped to a datamodel and return the mapping.

    Args:

        login (pycaol.Carol): Carol instance
        staging_name ([type]): staging name
        connector_name ([type]): connetor name
        connector_id ([type]): connetor id

    Usage:

    .. code:: python

        from pycarol import Carol
        from pycarol.functions import check_mapping
        carol = Carol()
        check_mapping(carol, staging_name='staging_name', connector_name='connector_name')

    Returns:

        list: list of mappings or None

    """
    conn = Connectors(login)

    if connector_id is None and connector_name is None:
        raise ValueError('Either connector_id or connector_name must be set.')
    connector_id = connector_id if connector_id else conn.get_by_name(connector_name)['mdmId']
    resp = conn.get_entity_mappings(
        connector_name=connector_name, staging_name=staging_name, connector_id=connector_id, 
        errors='ignore'
        )

    if isinstance(resp, dict):
        if resp['errorCode'] == 404 and 'Entity mapping not found' in resp['errorMessage']:
            return None
        else:
            raise ValueError(f'Error checking mapping {resp}')

    return resp

def resume_process(carol, staging_name, connector_name=None, connector_id=None, logger=None, delay=1):
    """Resume staging process (mappings or ETLs)

    Args:

        carol (pycarol.Carol): Carol instance
        staging_name (str): staging name
        connector_name (str): connector name
        connector_id (str): connetor id
        logger (logging.logger, optional): logger. Defaults to None.
        delay (int, optional): time to wait after resume processing. Defaults to 1.


    Usage:

    .. code:: python

        from pycarol import Carol
        from pycarol.functions import resume_process
        carol = Carol()
        resume_process(carol, staging_name='staging_name', connector_name='connector_name')

    Returns:

        list: list of mappings paused.
        
    """
    if logger is None:
        logger = logging.getLogger(carol.domain)

    conn = Connectors(carol)

    if connector_id is None and connector_name is None:
        raise ValueError('Either connector_id or connector_name must be set.')
    connector_id = connector_id if connector_id else conn.get_by_name(connector_name)['mdmId']
    # TODO Review this once we have mapping and ETLs in the same staging.
    # Play ETLs if any.
    resp = conn.play_etl(connector_id=connector_id, staging_name=staging_name, connector_name=connector_name)
    if not resp['success']:
        logger.error(f'Problem starting ETL {connector_name}/{staging_name}\n {resp}')
        raise ValueError(f'Problem starting ETL {connector_name}/{staging_name}\n {resp}')

    # Play mapping if any.
    mappings_list = check_mapping(carol, connector_name=connector_name, staging_name=staging_name, connector_id=connector_id)
    if mappings_list is not None:
        for mappings_ in mappings_list:
            conn.play_mapping(connector_name=connector_name, connector_id=connector_id, 
            entity_mapping_id=mappings_['mdmId'], process_cds=False, )

    # wait for mapping effect.
    time.sleep(delay)
    return mappings_list


def pause_etls(carol, etl_list, connector_name=None, connector_id=None, logger=None):
    """Pause ETLs from a connetor based on a list of ETLs.

    Args:

        carol (pycarol.Carol): Carol instance
        etl_list (list): ETLs to pause
        connector_name (str, optional): Connector Name. Defaults to None.
        connector_id (str, optional): Connector ID. Defaults to None.
        logger (logging.logger, optional): Logger. Defaults to None.

    Usage:

    .. code:: python

        from pycarol import Carol
        from pycarol.functions import resume_process
        carol = Carol()
        etl_list = ['staging1', 'staging2', 'staging3']
        pause_etls(carol, connector_name='rui', etl_list=etl_list,)

    Returns:

        list: list of paused ETLs

    """
    if logger is None:
        logger = logging.getLogger(carol.domain)

    conn = Connectors(carol)
    if connector_id is None and connector_name is None:
        raise ValueError('Either connector_id or connector_name must be set.')
    connector_id = connector_id if connector_id else conn.get_by_name(connector_name)['mdmId']

    r = {}
    for staging_name in etl_list:
        logger.debug(f'Pausing {staging_name} ETLs')
        r[staging_name] = conn.pause_etl(connector_name=connector_name, staging_name=staging_name)

    if not all(i['success'] for _, i in r.items()):
        logger.error(f'Some ETLs were not paused. {r}')
        raise ValueError(f'Some ETLs were not paused. {r}')

    return r

def delele_all_golden_data(carol, dm_name):
    """Delete golden files from a datamodel in all storages.

    Args:
        carol (pycarol.Carol): Carol instance
        dm_name (str): Data Model name

    Returns:
        list: list of tasks created
    """
    
    cds_CDSGolden = CDSGolden(carol)
    t = []
    dm_id = DataModel(carol).get_by_name(dm_name)['mdmId']
    task = cds_CDSGolden.delete_rejected(dm_id=dm_id, )

    t += [task]
    task = cds_CDSGolden.delete(dm_id=dm_id, )
    delete_golden(carol, dm_name)
    t += [task['taskId'], ]
    return t

def par_delete_golden(carol, dm_list, n_jobs=5):
    """
    Deletes golden files from a list of datamodels in parallel.

    Args:
        carol (pycarol.Carol): Carol instance
        dm_list (list): List of datamodels
        n_jobs (int, optional): Number of parallel jobs. Defaults to 5.

    Returns:
        list: list of tasks created
    """
    from joblib import Parallel, delayed
    tasks = Parallel(n_jobs=n_jobs)(delayed(delele_all_golden_data)(carol, i)
                                    for i in dm_list)
    return list(chain(*tasks))


def delete_staging_data(carol, staging_name, connector_name):
    """Delete a staging.

    Args:
        carol (pycarol.Carol): Login instance
        staging_name (str): Staging name
        connector_name (str): Connector name
    """ 

    t = []
    cds_ = CDSStaging(carol)
    task = cds_.delete(staging_name=staging_name,
                        connector_name=connector_name)
    t += [task['taskId'], ]
    return t

def par_delete_staging(carol, staging_list, connector_name, n_jobs=5):
    """
    Deletes staging files from a list of datamodels in parallel.

    Args:
        carol (pycarol.Carol): Login instance
        staging_list (list): List of datamodels
        connector_name (str): Connector name
        n_jobs (int, optional): Number of parallel jobs. Defaults to 5.

    Returns:
        list: list of tasks created
    """
    from joblib import Parallel, delayed
    tasks = Parallel(n_jobs=n_jobs)(delayed(delete_staging_data)(carol, i, connector_name)
                                    for i in staging_list)
    return list(chain(*tasks))



from pycarol.pipeline.utils.hash_versioning.test_hash_versioning import (
    equal_functions_list, different_functions_list
)

function_list = []
for a,b in equal_functions_list + different_functions_list:
    if a not in function_list:
        function_list.append(a)
    if b not in function_list:
        function_list.append(b)

from pycarol.pipeline.utils.hash_versioning import get_bytecode_tree, get_function_hash
def get_hash(f):
    name = f.__name__

    try:
        bytecode = get_bytecode_tree(f)
    except:
        bytecode = "FAIL"
    try:
        h = get_function_hash(f)
    except:
        h = "FAIL"
    
    return([name, bytecode, h])


def print_all():
    for f in function_list:
        name, bytecode, h = get_hash(f)
        print(name,h)


if __name__ == '__main__':
    print_all()


import pycarol


def test_dunder_version():

    assert pycarol.__version__


def test_dunder_temp_storage():

    assert pycarol.__TEMP_STORAGE__

def test_dunder_connector_id():

    assert pycarol.__CONNECTOR_PYCAROL__






from .data_models.data_models import DataModel

class Subscription:
    """
    Subscription from carol.

    Args:

        Carol object
            Carol object.
    """

    def __init__(self, carol):
        self.carol = carol

    def get_all(self, offset=0, page_size=1000, sort_by=None, sort_order='ASC'):
        """
        Get all subscription information.

        Args:
            offset: `int`, default 0
            Offset for pagination. Only used when `scrollable=False`
        page_size: `int`, default 100
            Number of records downloaded in each pagination. The maximum value is 1000
        sort_order: `str`, default 'ASC'
            Sort ascending ('ASC') vs. descending ('DESC').
        sort_by: `str`,  default `None`
            Name to sort by.

        Returns: list
             list of subscription

        """
        query_params = {"offset": offset, "pageSize": page_size, "sortOrder": sort_order,
                        "sortBy": sort_by}
        return self.carol.call_api('v2/subscription', method='GET' ,params=query_params)['hits']

    def create_edit(self, subscription_list):
        """
        Subscribe to a entity or edit the subscription. Attention: Today the "pk"" for a subscription is
        `connector_id` + `dm_id`. It means that is possible to have two subscriptions for the same datamodel
        using different connectors.

        Args:
            subscription_list:
                list of subscriptions to create/edit.

        Returns: dict
            carol response.

        """


        return self.carol.call_api('v2/subscription/batch', method='POST', data=subscription_list)

    def pause(self, subscription_id):
        """
        Pause a subscription

        Args:
            subscription_id:
                if og the subscription

        Returns: dict
            carol response.

        """
        return self.carol.call_api(f'v2/subscription/{subscription_id}/pause', method='POST')


    def clear(self, subscription_id):
        """
        Clear the pending records from a subscription

        Args:
            subscription_id:
                if og the subscription

        Returns: dict
            carol response.

        """
        return self.carol.call_api(f'v2/subscription/{subscription_id}/clear', method='POST')

    def play(self, subscription_id):
        """
        Play a subscription

        Args:
            subscription_id:
                if og the subscription

        Returns: dict
            carol response.

        """
        return self.carol.call_api(f'v2/subscription/{subscription_id}/play', method='POST')


    def get_by_id(self, subscription_id):
        """
        Get a subscription by ID

        Args:
            subscription_id:
                if og the subscription

        Returns: dict
            carol response.

        """
        return self.carol.call_api(f'v2/subscription/{subscription_id}', method='GET')

    def get_info(self, subscription_id):
        """
        Get information for a subscription by ID

        Args:
            subscription_id:
                if og the subscription

        Returns: dict
            carol response.

        """
        return self.carol.call_api(f'v2/subscription/{subscription_id}/details', method='GET')

    def get_dm_subscription(self, dm_name=None, dm_id=None):
        """
        Get DataModel subscription information.

        Args:
            dm_name: `str` default `None`
                DataModel name
            dm_id: `str` default `None`
                DataModel id

        Returns: dict
            carol response.

        """

        if dm_name is not None:
            resp = DataModel(self.carol).get_by_name(dm_name)
            dm_id = resp['mdmId']
        elif dm_id is None:
            raise ValueError('Either `dm_name` or `dm_id` must be set.')

        DataModel(self.carol)
        return self.carol.call_api(f'v2/subscription/template/{dm_id}', method='GET')


















import types
import time
from .PwdAuth import PwdAuth


class PwdKeyAuth(PwdAuth):
    """

    Args:

        access_token: `str`
            Carol access token created using the user and password.
        lazy_login: `bool` default `False`
            It will not validate the `access_token` when starting the instance.
            This is used to speedup requests, when the user knows that this token is valid. Using this option, one lose
            all refresh token capability.
    """

    def __init__(self, access_token, lazy_login=False):
        self.access_token = access_token
        self._token = None
        self.carol = None
        self.connector_id = None
        self.lazy_login = lazy_login


    def login(self, carol):
        """

        Args:

            carol: pycarol.Carol
                Carol() instance.

        Returns:
            None

        """
        self.carol = carol

        if self.lazy_login:
            self._token = types.SimpleNamespace()
            self._token.access_token = self.access_token
            self._token.refresh_token = None
            self._token.expiration = float('inf')
            return


        resp = self.carol.call_api(f'v2/oauth2/token/{self.access_token}', method='GET', auth=False,
                                   content_type='application/x-www-form-urlencoded')

        self._set_token(resp)
        if self.carol.verbose:
            print("Token: {}".format(self._token.access_token))

    def _is_token_expired(self):
        if self._token is None:
            self.login()
            return True

        if self._token.expiration == 0:
            return False

        now = time.time() * 1000
        # Adds 1 min buffer
        expiry = self._token.expiration - 60000

        return now > expiry



"""
Carol app funtionalities.


"""

import zipfile, io, os
from .utils.deprecation_msgs import _deprecation_msgs
from .utils.miscellaneous import zip_folder
import tempfile

class Apps:
    """
    Carol App instance.

        Args:

            carol: `pycarol.Carol`
                Carol() instance


    """

    def __init__(self, carol):
        """
        Initialize Class

        Args:

            carol: `pycarol.Carol`
                Carol() instance
        """
        self.carol = carol

    def _build_query_params(self, entity_space="PRODUCTION", offset=0, page_size=100, sort_by=None, sort_order=None):
        if sort_by is None:
            return {"offset": offset, "pageSize": page_size, "entitySpace": entity_space}
        else:
            return {"offset": offset, "pageSize": page_size, "sortOrder": sort_order,
                    "sortBy": sort_by, "entitySpace": entity_space}

    def _define_current_run(self, query):
        self.current_app_id = query.get('mdmId')
        self.current_app_name = query.get('mdmName')
        self.current_app = {self.current_app_name: query}

    def all(self, entity_space='PRODUCTION', page_size=50, offset=0, sort_order='ASC', sort_by=None):
        """
        Get all app information in this environment.

        Args:

            entity_space: `str` default `PRODUCTION`
                Space to get the app from. Possible values

                    1. PRODUCTION: For production
                    2. WORKING: For Draft apps.

        offset: `int`, default 0
            Offset for pagination. Only used when `scrollable=False`
        page_size: `int`, default 100
            Number of records downloaded in each pagination. The maximum value is 1000
        sort_order: `str`, default 'ASC'
            Sort ascending ('ASC') vs. descending ('DESC').
        sort_by: `str`,  default `None`
            Name to sort by.

        Returns: List of Dicts.
            All apps json definition

        """

        query_string = self._build_query_params(offset=offset, page_size=page_size, entity_space=entity_space,
                                                sort_by=sort_by, sort_order=sort_order)
        query = self.carol.call_api('v1/tenantApps', method='GET', params=query_string)

        self.all_apps = {app['mdmName']: app for app in query['hits']}
        self.total_hits = query["totalHits"]
        return self.all_apps

    def get_by_name(self, app_name, entity_space='PRODUCTION'):
        """
        Get app information by app name

        Args:

            app_name: `str`
                App name.
            entity_space: `str` default `PRODUCTION`
                Space to get the app information from. Possible values

                    1. PRODUCTION: For production
                    2. WORKING: For Draft apps.

        Returns: `dict`
            App info json definition.

        """

        query_string = {"entitySpace": entity_space}
        query = self.carol.call_api(f'v1/tenantApps/name/{app_name}', method='GET', params=query_string)

        self._define_current_run(query)

        return query

    def get_by_id(self, app_id, entity_space='PRODUCTION'):
        """
        Get app information by ID

        Args:

            app_id: `str`
                App id.
            entity_space: `str` default `PRODUCTION`
                Space to get the app information from. Possible values

                    1. PRODUCTION: For production
                    2. WORKING: For Draft apps.

        Returns: `dict`
            App info json definition.

        """

        query_string = {"entitySpace": entity_space}
        query = self.carol.call_api(f'v1/tenantApps/{app_id}', method='GET', params=query_string)

        self._define_current_run(query)

        return query

    def get_settings(self, app_name=None, app_id=None, entity_space='PRODUCTION', check_all_spaces=False):
        """
        Get settings from app

        Settings are the settings available in Carol's UI.

        Args:

            app_name: `str` default `None`
                App name, if None will get from  `app_id`
            app_id: `str` default `None`
                App id. Either app_name or app_id must be set.
            entity_space: `str` default `PRODUCTION`
                Space to get the app settings from. Possible values

                    1. PRODUCTION: For production
                    2. WORKING: For Draft apps.

            check_all_spaces: `bool` default `False`
                Check all entity spaces.

        Returns: `dict`
            Dictionary with the settings
        """

        assert app_name or app_id or self.carol.app_name

        if app_id is not None:
            self.get_by_id(app_id, entity_space)
        elif app_name is not None:
            self.get_by_name(app_name, entity_space)
        else:
            self.get_by_name(self.carol.app_name, entity_space)

        query_string = {"entitySpace": entity_space, "checkAllSpaces": check_all_spaces}

        query = self.carol.call_api(f'v1/tenantApps/{self.current_app_id}/settings', method='GET',
                                    params=query_string)

        self.app_settings = {}
        self.full_settings = {}

        if not isinstance(query, list):
            query = [query]
        for query_list in query:
            self.app_settings.update({i['mdmName']: i.get('mdmParameterValue')
                                      for i in query_list.get('mdmTenantAppSettingValues',{})})
            self.full_settings.update({i['mdmName']: i for i in query_list.get('mdmTenantAppSettingValues',{})})

        return self.app_settings

    def download_app(self, app_name=None, app_version=None, carolappname=None, carolappversion=None,
                     file_path='carol.zip', extract=False):
        """
        Download App artifacts.

        Args:

            app_name: `str` default `None`
                Carol app name. It will overwrite the app name used in Carol() initialization.
            app_version: `str`
                App Version
            carolappname: `str`
                App Name. Deprecated. Use app_name
            carolappversion: `str`
                App Version. Deprecated. Use app_version
            file_path:  `os.PathLike`
                Path to save the zip file.
            extract: `bool` default `False`
                Either extract the zip files or not.

        Returns: None

        """

        if carolappname is not None:
            app_name = carolappname
            _deprecation_msgs("`carolappname` is deprecated use `app_name`.")
        if carolappversion is not None:
            app_version = carolappversion
            _deprecation_msgs("`carolappversion` is deprecated use `app_version`.")

        if app_name is None:
            app_name = self.carol.app_name

        if app_version is None:
            raise ValueError('app_version must be set.')

        url = f'v1/carolApps/download/{app_name}/version/{app_version}'

        r = self.carol.call_api(url, method='GET', stream=True, downloadable=True)

        if extract:
            r = zipfile.ZipFile(io.BytesIO(r.content))
            r.extractall(file_path)
        else:
            with open(file_path, 'wb') as out:  ## Open temporary file as bytes
                out.write(io.BytesIO(r.content).read())

    def get_manifest(self, app_name=None):
        """

        Args: `str`

            app_name: `str` default `None`
                Carol app name. It will overwrite the app name used in Carol() initialization.

        Returns: `dict`
            Dictionary with the manifest file.

        """

        if app_name is None:
            app_name = self.carol.app_name

        url = f'v1/tenantApps/manifest/{app_name}'

        r = self.carol.call_api(url, method='GET')
        return r

    def edit_manifest(self, manifest, app_name=None):
        """

        Args:

            manifest: `dict`
                Dictionary with the manifest
            app_name: `str` default `None`
                Carol app name. It will overwrite the app name used in Carol() initialization.

        Returns: `dict`
            {"success": True}

        """

        if app_name is None:
            app_name = self.carol.app_name

        url = f'v1/tenantApps/manifest/{app_name}'

        r = self.carol.call_api(url, method='PUT', data=manifest)
        return r

    def get_git_process(self, app_name=None):
        """
        Get Git processes definid in the manifest file.

        Args:

            app_name: `str` default `None`
                Carol app name. It will overwrite the app name used in Carol() initialization.

        Returns: List of Dicts

        """

        if app_name is None:
            app_name = self.carol.app_name

        response = self.carol.call_api(path=f'v1/compute/{app_name}/getProcessesByGitRepo',
                                       method='POST')

        return response

    def build_docker_git(self, git_token, app_name=None, ):
        """
        Build App image using manifest definition.

        Args:
            git_token: `str`
                Git token to be used to pull the files.

            app_name: `str` default `None`
                Carol app name. It will overwrite the app name used in Carol() initialization.

        Returns:
            List of task ids.
        """

        if app_name is None:
            app_name = self.carol.app_name

        manifest = self.get_git_process(app_name)

        self._assert_manifest_fields(manifest)

        tasks = []
        for build in manifest:
            docker_name = build['dockerName']
            docker_tag = build['dockerTag']
            instance_type = build['instanceType']
            git_token = git_token
            params = {
                'dockerName': docker_name,
                'tagName': docker_tag,
                'gitToken': git_token,
                'instanceType': instance_type,
            }
            response = self.carol.call_api(path=f'v1/compute/{app_name}/buildGitDocker',
                                           method='POST', params=params)

            tasks.append(response)

        return tasks

    @staticmethod
    def _assert_manifest_fields(manifest):
        """
        Assert that the fields needed to build the image exist.
        Args:
            manifest: `list of dict`
                list of docker definition in the manifest file.

        Returns:
            None
        """

        fields = {'dockerName', 'dockerTag', 'instanceType'}
        for build in manifest:
            set_diff = fields - set(build)
            if len(set_diff) >= 1:
                raise ValueError(f'Missing docker definition {set_diff}')

    def update_setting_values(self, settings, app_name=None):
        """
        Change Settings values in Carol.

        Args:
            settings: `dict`
                dict with settings:
                {"param1": "value1", "param2": "value2"}
            app_name: `str` default None
                App name to change the settings.

        Returns: `dict`
            Carol Response.


        """

        if app_name is None:
            app_name = self.carol.app_name

        if not isinstance(settings, dict):
            ValueError(f"settings should be a dictionary,")

        app_id = self.get_by_name(app_name)['mdmId']
        settings_id = self._get_app_settings_config(app_id=app_id)['mdmId']
        data = [{"mdmName": i, "mdmParameterValue": j} for i, j in settings.items()]

        return self.carol.call_api(path=f'v1/tenantApps/{app_id}/settings/{settings_id}?publish=true', method='PUT',
                                   data=data)

    def _get_app_settings_config(self, app_name=None, app_id=None):

        if app_id is not None:
            pass
        elif app_name is None:
            app_name = self.carol.app_name
            app_id = self.get_by_name(app_name)['mdmId']

        return self.carol.call_api(path=f'v1/tenantApps/{app_id}/settings', )

    def start_app_process(self, process_name, app_name=None, ):

        """
        Start a carol process by process name.

        Args:
            process_name: `str`
                Process name.
            app_name: `str` default None
                App name to change the settings.

        Returns: `dict`
            task information.

        """

        if app_name is None:
            app_name = self.carol.app_name

        app_id = self.get_by_name(app_name)['mdmId']
        process_id = self.get_processes_info(app_name=app_name)["mdmId"]
        return self.carol.call_api(f"v1/tenantApps/{app_id}/aiprocesses/{process_id}/execute/{process_name}",
                                   method='POST')

    def get_processes_info(self, app_name=None, entity_space='WORKING', check_all_spaces=True):
        """
        Get app processes information.
        Args:
            app_name: `str` default None
                App name to change the settings.
            entity_space: `str` default `WORKING`
                WORKING or PRODUCTION
            check_all_spaces: `bool`
                Check all spaces.

        Returns:
            Process informations
        """

        params = {"entitySpace": entity_space, "checkAllSpaces": check_all_spaces}
        if app_name is None:
            app_name = self.carol.app_name
        app_id = self.get_by_name(app_name)['mdmId']
        return self.carol.call_api(f"v1/tenantApps/{app_id}/aiprocesses", method='GET', params=params)

    def get_subscribable_carol_apps(self):
        """
        Find all available apps to install in this env.

        Returns: `list`
            list of apps.

        """

        return self.carol.call_api("v1/tenantApps/subscribableCarolApps", method='GET')['hits']

    def install_carol_app(self, app_name=None, app_version=None, connector_group=None, publish=True):

        """
        Install a carol app in an env.

        Args:
            app_name: `str` default None
                App name to change the settings.
            app_version: `str` default None
                App version to install. If not specified, it will install the most recent.
            connector_group: `str` default None
                Connector Group to install.
            publish: `bool` default True
                If publish the update.

        Returns:
            Carol task.

        """

        if app_name is None:
            app_name = self.carol.app_name

        to_install = self.get_subscribable_carol_apps()
        to_install = [i for i in to_install if i["mdmName"] == app_name]


        if app_version == None:
            to_install = sorted(to_install, key=lambda x: x['mdmAppVersion'])
        else:
            to_install = [i for i in to_install if i["mdmAppVersion"] == app_version]

        if to_install:
            to_install = to_install[0]
        else:
            return

        to_install_id = to_install['mdmId']

        updated = self.carol.call_api(f"v1/tenantApps/subscribe/carolApps/{to_install_id}", method='POST')
        params = {"publish": publish, "connectorGroup": connector_group}
        return self.carol.call_api(f"v1/tenantApps/{updated['mdmId']}/install", method='POST', params=params)

    def get_app_details(self, app_name=None, entity_space='PRODUCTION'):
        """

        Find all information about an app. This will fetch information about connector groups, AI process, descriptions,
        data models, etc.

        Args:
            app_name: `str` default None
                App name to change the settings.
            entity_space: `str` default `PRODUCTION`
                WORKING or PRODUCTION

        Returns: `dict`
            Carol response.
        """

        if app_name is None:
            app_name = self.carol.app_name

        #check if it exists as a subscribable apps
        to_install = self.get_subscribable_carol_apps()
        to_install = [i for i in to_install if i["mdmName"] == app_name]

        if to_install:
            to_install = to_install[0]
            app_id = to_install['mdmId']
        else:
            # try installed app.
            app_id = self.get_by_name(app_name)['mdmCarolAppId']

        params = {"entitySpace": entity_space}
        return self.carol.call_api(f"v1/carolApps/{app_id}/details", method='GET', params=params)

    def upload_file(self, filepath,  app_name=None,):
        """
        Upload the app artifacts

        Args:
            filepath: `str`
                Folder path with artifacts (manifest.json or site)

            app_name: `str` default None
                App name

        Returns: `dict`
            Carol response.
        """

        if app_name is None:
            app_name = self.carol.app_name

        app_id = self.get_by_name(app_name)['mdmCarolAppId']

        if filepath.endswith(('.zip', '.json')):
            file = filepath
        elif os.path.isdir(filepath):
            file = zip_folder(filepath)
        uri = f'v1/carolApps/{app_id}/files/upload'

        with open(file, 'rb') as f:
            files = {'file': f}
            r = self.carol.call_api(path=uri, method='POST', files=files, content_type=None)

        return r







"""


"""

import luigi
from luigi import parameter
from luigi.task import flatten
from luigi.parameter import ParameterVisibility
from pycarol.pipeline.targets import PickleTarget
from pycarol.utils.miscellaneous import Hashabledict
import logging
import warnings
import types


logger = logging.getLogger('luigi-interface')
logger.setLevel(logging.INFO)


class Task(luigi.Task):
    """
    This is the base class of all pipelines.

    This Task is an extension of :py:class:`luigi.Task` task. We added some functionalities to it:
        1. No need to define the method 'output` and manually save the output`. Just return the object to be saved
        in the `easy_run` method.
        2. No need to define the requirements using the `require` method. It is possible to use
        :py:class:`pycarol.pipeline.inherit_list` and :py:class:`pycarol.pipeline.inherit_dict` decorators.
        3. No need to load manually the input task. `easy_run` method will receive a list of all the requirements loaded
        in the same order they were defined in :py:class:`pycarol.pipeline.inherit_list` or the dictionary if it was used
        :py:class:`pycarol.pipeline.inherit_dict`


    There are two ways to execute the computation of a task.
        1. Implementing the method: * :py:meth:`easy_run`
        2. Defining either the class variable `task_function` or `task_notebook`. Refer to main documentation
        for more details.

    """
    TARGET_DIR = './TARGETS/'
    target_type = PickleTarget
    is_cloud_target = None
    requires_list = []
    requires_dict = {}

    task_function = None
    task_notebook = None
    easy_run = None
    version = '0.0.0'
    metadata = {}

    def get_task_address(self):
        if self.task_notebook:
            return self.task_notebook
        else:
            return None

    def buildme(self, local_scheduler=True, **kwargs):
        luigi.build([self, ], local_scheduler=local_scheduler, **kwargs)

    def _file_id(self):
        # returns the output default file identifier
        return luigi.task.task_id_str(self.get_task_family(), self.to_str_params(only_significant=True))

    def requires(self):
        if len(self.requires_list) > 0:
            result_list = []
            for t in self.requires_list:
                fixed_params = {}
                if type(t) is tuple:
                    fixed_params = t[1]
                    t = t[0]
                task_instance = self.clone(t, **fixed_params)
                result_list.append(task_instance)
            return result_list
        elif len(self.requires_dict) > 0:
            result_dict = {}
            for k, t in self.requires_dict.items():
                fixed_params = {}
                if type(t) is tuple:
                    fixed_params = t[1]
                    t = t[0]
                task_instance = self.clone(t, **fixed_params)
                result_dict.update({k: task_instance})
            return result_dict
        else:
            return []

    def output(self):
        if hasattr(self,'TARGET'):  # Check for deprecated use
            warnings.warn('TARGET is being replaced with target_type.', DeprecationWarning)
            return self.TARGET(self)

        return self.target_type(self)

    def load(self, **kwargs):
        return self.output().load(**kwargs)

    def load_metadata(self):
        return self.output().load_metadata()

    def remove(self):
        self.output().remove()
        self.output().remove_metadata()

    def save(self):
        self.output().dump(self.output_object)
        self.output().dump_metadata(self.metadata())

    def metadata(self):
        metadata = dict()
        metadata['hash_version'] = self.hash_version()
        metadata['version'] = self.version
        metadata['params'] = self.get_execution_params(only_significant=False, only_public=True)
        return metadata

    def run(self):
        if self.easy_run:
            inputs = self.function_inputs()
            self.output_object = self.easy_run(inputs)

            if isinstance(self.output_object, types.GeneratorType):
                # Allowing dynamic tasks.
                # https://luigi.readthedocs.io/en/stable/tasks.html#dynamic-dependencies
                while True:
                    try:
                        new_requires = next(self.output_object)
                    except StopIteration as e:
                        # We expect the result of the task in the return of the iterator.
                        self.output_object = e.value
                        break
                    # From here, Luigi assumes and generate the new dependencies.
                    yield new_requires
            self.save()
            del self.output_object  # after dump, free memory

        elif self.task_function:
            inputs = self.function_inputs()
            if not isinstance(inputs,list):
                raise NotImplementedError(
                    f"In task_function mode, inputs should be list, not {type(inputs)}"
                    )
            params = self.get_execution_params(only_significant=True)
            assert hasattr(self.task_function,'__func__'), "We need unbound method"
            f = self.task_function.__func__
            self.output_object = f(*inputs, **params)
            self.save()
            del self.output_object  # after dump, free memory

        elif self.task_notebook:
            import papermill
            #TODO: create output folder
            papermill.execute_notebook(
                self.task_notebook,
                # f"executed_notebook/{self.task_notebook}",
                "/dev/null",
                parameters=dict(),
            )
            # self.save is called inside notebook

        else:
            raise SyntaxError("One of [easy_run, task_function, task_notebook] "
                              "should be defined")

    def function_inputs(self):
        if isinstance(self.input(), list):
            function_inputs = [input_i.load(
                **self.load_input_params(input_i)) if self.load_input_params(
                input_i) else input_i.load() for input_i in self.input()]
        elif isinstance(self.input(), dict):
            function_inputs = {i: (input_i.load(
                **self.load_input_params(input_i)) if self.load_input_params(
                input_i) else input_i.load()) for i, input_i in
                               self.input().items()}
        else:
            raise NotImplementedError(f"input should be either list or dict. "
                                      f"received {type(self.input())}")
        return function_inputs


    def hash_version(self,):
        """ Returns the hash of the task considering only function, not the parameters."""
        from ..utils.hash_versioning import get_function_hash
        if not self.task_function:
            warnings.warn(
                "hash versioning only works in task_function mode. "\
                "It will return dummy hash code",SyntaxWarning
                )
            return 0
        else:
            try:
                return get_function_hash(self.task_function, ignore_not_implemented=True)
            except:
                return 0

    @classmethod
    def get_param_values(cls, params, args, kwargs):
        """
        This method was changed from the original version to allow execution of a task
        with extra parameters. the original one, raises an exception. now, we print 
        that exception in this version we do not raise neither print it.

        Get the values of the parameters from the args and kwargs.
        :param params: list of (param_name, Parameter).
        :param args: positional arguments
        :param kwargs: keyword arguments.
        :returns: list of `(name, value)` tuples, one for each parameter.
        
        """
        result = {}

        params_dict = dict(params)

        task_family = cls.get_task_family()

        # In case any exceptions are thrown, create a helpful description of how the Task was invoked
        # TODO: should we detect non-reprable arguments? These will lead to mysterious errors
        exc_desc = '%s[args=%s, kwargs=%s]' % (task_family, args, kwargs)

        # Fill in the positional arguments
        positional_params = [(n, p) for n, p in params if p.positional]
        for i, arg in enumerate(args):
            if i >= len(positional_params):
                raise parameter.UnknownParameterException(
                    '%s: takes at most %d parameters (%d given)' % (exc_desc, len(positional_params), len(args)))
            param_name, param_obj = positional_params[i]
            result[param_name] = param_obj.normalize(arg)

        # Then the keyword arguments
        for param_name, arg in kwargs.items():
            if param_name in result:
                raise parameter.DuplicateParameterException(
                    '%s: parameter %s was already set as a positional parameter' % (exc_desc, param_name))
            if param_name not in params_dict:
                # this is the difference between our and luigi's implementations.
                # raise parameter.UnknownParameterException('%s: unknown parameter %s' % (exc_desc, param_name))
                continue

            result[param_name] = params_dict[param_name].normalize(arg)

        # Then use the defaults for anything not filled in
        for param_name, param_obj in params:
            if param_name not in result:
                if not param_obj.has_task_value(task_family, param_name):
                    raise parameter.MissingParameterException(
                        "%s: requires the '%s' parameter to be set" % (exc_desc, param_name))
                result[param_name] = param_obj.task_value(task_family, param_name)

        def list_to_tuple(x):
            """ Make tuples out of lists and sets to allow hashing """
            if isinstance(x, list) or isinstance(x, set):
                return tuple(x)
            else:
                return x

        # Sort it by the correct order and make a list
        return [(param_name, list_to_tuple(result[param_name])) for param_name, param_obj in params]

    def get_execution_params(self, only_significant=False, only_public=True):
        """
        Get params values.


        """
        params_str = {}
        params = dict(self.get_params())
        for param_name, param_value in self.param_kwargs.items():
            if (((not only_significant) or params[param_name].significant)
                    and ((not only_public) or params[param_name].visibility == ParameterVisibility.PUBLIC)
                    and params[param_name].visibility != ParameterVisibility.PRIVATE):

                #TODO: Should we save the :class: luigi.Parameter itself?
                params_str[param_name] = param_value

        return params_str


    def load_input_params(self, input_target):
        """
        Overwrite this if need to pass parameters when loading a requirement.

        :param input_target:
            Target that will be loaded.
        :return: `Dict`
            Return a dict with key/value parameters to be passed to Target.load()
        """
        return {}

#TODO: remove either WrapperTask or Dummy Target
class WrapperTask(Task):
    """
    Use for tasks that only wrap other tasks and that by definition are done if all their requirements exist.
    """
    def run(self):
        pass

    def complete(self):
        return all(r.complete() for r in flatten(self.requires()))

    def output(self):
        return self.input()


def set_attributes(task_to_inherit, task_that_inherits):
    fixed_params = {}

    if type(task_to_inherit) is tuple:
        fixed_params = task_to_inherit[1]
        task_to_inherit = task_to_inherit[0]

    for param_name, param_obj in task_to_inherit.get_params():
        if param_name in fixed_params:  # do not inherit fixed params
            continue

        # Check if the parameter exists in the inheriting task
        if not hasattr(task_that_inherits, param_name):
            # If not, add it to the inheriting task
            setattr(task_that_inherits, param_name, param_obj)
    return task_that_inherits


class inherit_list(object):
    # http://blog.thedigitalcatonline.com/blog/2015/04/23/python-decorators-metaprogramming-with-style/

    def __init__(self, *task_to_inherit_list):
        self.requires_list = list(task_to_inherit_list)
        # next, we use hashable dict in local task params to support pipeline viewer
        for i,v in enumerate(self.requires_list):
            if isinstance(v,tuple):
                task, params = v
                assert issubclass(task,Task)
                assert isinstance(params,dict)
                self.requires_list[i] = ( task, Hashabledict(params) )

    def __call__(self, task_that_inherits):
        task_that_inherits.requires_list = self.requires_list
        for task_to_inherit in task_that_inherits.requires_list:
            # Get all parameter objects from the underlying task
            task_that_inherits = set_attributes(task_to_inherit, task_that_inherits)

        return task_that_inherits


class inherit_dict(object):
    #TODO: hash versioning is not compatible with inherit_dict
    def __init__(self, **task_to_inherit_dict):
        self.requires_dict = task_to_inherit_dict

    def __call__(self, task_that_inherits):
        task_that_inherits.requires_dict = self.requires_dict

        for key, task_to_inherit in task_that_inherits.requires_dict.items():
            # Get all parameter objects from the underlying task
            task_that_inherits = set_attributes(task_to_inherit, task_that_inherits)
        return task_that_inherits



import luigi
import pandas as pd
import os
import joblib
import warnings

class LocalTarget(luigi.LocalTarget):
    FILE_EXT = ''
    is_tmp = False

    def __init__(self, task, path=None, *args, **kwargs):
        os.makedirs(task.TARGET_DIR, exist_ok=True)
        namespace = task.get_task_namespace()
        self.task = task  # keeping task object in target object will make
        # load_inputs_params easiers
        if path is None:
            file_id = task._file_id()
            ext = '.' + self.FILE_EXT
            file_id = file_id.split(namespace+'.')[-1]  #this will prevent to copy all the module path to the name of the file.
            path = os.path.join(task.TARGET_DIR, namespace, file_id + ext)
        super().__init__(path=path, *args, **kwargs)

    def dump_metadata(self, metadata: dict,  *args, **kwargs):
        warnings.warn("dump_metadata not implemented in LocalTarget")

    def load_metadata(self, *args,**kwargs):
        """Should return a dict."""
        warnings.warn("load_metadata not implemented in LocalTarget")
        return {}

    def remove_metadata(self, *args,**kwargs):
        warnings.warn("remove_metadata not implemented in LocalTarget")

    def get_metadata_path(self, *args,**kwargs):
        return f"{self.path}.metadata"        

class CDSTarget(LocalTarget):
    """ A target that works both locally and on Carol Data Storage, based on env parameter CLOUD_TARGET

    In order to use Carol Data Storage Targets, env or files configuration should allow Carol authentication
    with no parameters, Carol().
    If more than one tenant are used in same session, luigi parameter "tenant" should exist.

    """
    login_cache = None
    tenant_cache = None
    storage_cache = None

    def __init__(self, task, *args, **kwargs):
        super().__init__(task, *args, **kwargs)

        env_is_cloud_target = os.environ.get('CLOUD_TARGET', 'false').lower() == 'true'
        self._is_cloud_target = task.is_cloud_target or env_is_cloud_target
        env_tenant = os.environ.get('CAROLTENANT',None)
        if hasattr(task,'tenant'):
            tenant = task.tenant
        else:
            tenant = env_tenant
        if self._is_cloud_target:
            from pycarol.carol import Carol
            from pycarol.storage import Storage

            # We CANNOT cache the storage with GCP because the GCP API is not thread safe and would result in SSL errors
            # if luigi is using more than 1 worker
            #TODO: login_cache will not work in multiprocess mode. disable login_cache ?
            if (CDSTarget.login_cache) and (CDSTarget.tenant_cache == tenant):
                self.login = CDSTarget.login_cache
            else:
                self.login = Carol()
                CDSTarget.login_cache = self.login
                CDSTarget.tenant_cache = tenant 
            # Storage var needs to be always created per Target
            self.storage = Storage(self.login)
    
            namespace = task.get_task_namespace()
            file_id = task._file_id()
            file_id = file_id.split(namespace+'.')[-1] #this will prevent to copy all the module path to the name of the file.
            self._local_path = self.path    # Save a copy of the local path, before modifying it: This is useful when
                                            # the target needs to use a local path
            self.path = os.path.join('pipeline', namespace, "{}.{}".format(file_id, self.FILE_EXT))
            self.log_path = os.path.join('pipeline',namespace, "{}_log.pkl".format(file_id))

    def dump_metadata(self, metadata:dict, *args,**kwargs):
        if self._is_cloud_target:
            assert isinstance(metadata, dict)
            self.storage.save(self.get_metadata_path(), metadata, format='joblib', cache=False)
        else:
            super().dump_metadata(metadata, *args, **kwargs)

    def load_metadata(self, *args, **kwargs):
        """Should return a dict."""
        if self._is_cloud_target:
            metadata = self.storage.load(self.get_metadata_path(), format='joblib', cache=False)
            assert isinstance(metadata,dict), f"metadata is type {type(metadata)}"
            return metadata
        else:
            return super().load_metadata(*args,**kwargs)

    def remove_metadata(self,*args,**kwargs):
        if self._is_cloud_target:
            self.storage.delete(self.get_metadata_path())
        else:
            return super().load_metadata(*args,**kwargs)


    def load(self, *args, **kwargs):
        if self._is_cloud_target:
            return self.load_cds(*args, **kwargs)
        return self.load_local(*args, **kwargs)

    def dump(self, *args, **kwargs):
        if self._is_cloud_target:
            return self.dump_cds(*args, **kwargs)
        return self.dump_local(*args, **kwargs)

    def remove(self, *args, **kwargs):
        if self._is_cloud_target:
            return self.remove_cds(*args, **kwargs)
        return self.remove_local(*args, **kwargs)

    def exists(self, *args, **kwargs):
        if self._is_cloud_target:
            return self.exists_cds(*args, **kwargs)
        return self.exists_local(*args, **kwargs)

    def load_local(self, *args, **kwargs):
        return super().load(*args, **kwargs)

    def dump_local(self, *args, **kwargs):
        return super().dump(*args, **kwargs)

    def exists_local(self, *args, **kwargs):
        return super().exists()

    def remove_local(self, *args, **kwargs):
        return super().remove()

    def remove_cds(self, *args, **kwargs):
        self.storage.delete(self.path)

    def exists_cds(self, *args, **kwargs):
        return self.storage.exists(self.path)
    

class FileTarget(CDSTarget):
    """
    This target operates with filepaths.
    easy_run should return a filepath for a local temporary file. This file will be removed after been sent to Carol.
    When loading this target, the file is copied from Carol to a local file. On easy_run we receive the local filepath.
    Important note: when loading the target, its local copy will not be automatically removed.
    """

    FILE_EXT = 'file'

    # TODO NOW: Define how it should be done locally

    def load_cds(self):
        return self.storage.load(self.path, format='file', cache=False)

    def dump_cds(self, tempfile_path):
        self.storage.save(self.path, tempfile_path, format='file', cache=False)
        os.remove(tempfile_path)


class PickleTarget(CDSTarget):
    FILE_EXT = 'pkl'

    def load_cds(self):
        return self.storage.load(self.path, format='joblib', cache=False)

    def dump_cds(self, function_output):
        self.storage.save(self.path, function_output, format='joblib', cache=False)

    def load_local(self):
        return joblib.load(self.path)

    def dump_local(self, function_output):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        joblib.dump(function_output, self.path)


class ParquetTarget(CDSTarget):
    FILE_EXT = 'parquet'

    def load_cds(self, **kwargs):
        return self.storage.load(self.path, format='joblib', cache=True, parquet=True, **kwargs)

    def dump_cds(self, function_output):
        self.storage.save(self.path, function_output, format='joblib', cache=False, parquet=True)

    def load_local(self, **kwargs):
        return pd.read_parquet(self.path, **kwargs)

    def dump_local(self, function_output):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        function_output.to_parquet(self.path)


class KerasTarget(CDSTarget):
    FILE_EXT = 'h5'

    def load_cds(self):
        from keras.models import load_model
        local_path = self.storage.load(self.path, format='file')
        return load_model(local_path)

    def dump_cds(self, model):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        model.save(self.path)
        self.storage.save(self.path, self.path, format='file')

    def load_local(self):
        from keras.models import load_model
        return load_model(self.path)

    def dump_local(self, model):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        model.save(self.path)


class PytorchTarget(CDSTarget):
    FILE_EXT = 'pth'

    def load_cds(self):
        import torch
        local_path = self.storage.load(self.path, format='file')
        return torch.load(local_path)

    def dump_cds(self, model_state_dict):
        import torch
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        torch.save(model_state_dict, self.path)
        self.storage.save(self.path, self.path, format='file')

    def load_local(self):
        import torch
        return torch.load(self.path)

    def dump_local(self, model_state_dict):
        import torch
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        torch.save(model_state_dict, self.path)


class DummyTarget:

    def __init__(self, fixed_output=None, *args, **kwargs):
        self.fixed_output = fixed_output

    def exists(self):
        return True

    def complete(self):
        return all(r.complete() for r in flatten(self.requires()))

    def load(self):
        return self.fixed_output

    def dump(self, model):
        pass

    def remove(self):
        pass


class JsonTarget(CDSTarget):
    FILE_EXT = 'json'

    # TODO NOW: Define how to do that on the CDS

    def load_local(self):
        return pd.read_json(self.path)

    def dump_local(self, function_output):
        #TODO: json only works for dataframe
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        function_output.to_json(self.path)


class FeatherTarget(CDSTarget):
    FILE_EXT = 'feather'

    # TODO NOW: Define how to do that on the CDS

    def load_local(self):
        import feather
        return feather.read_dataframe(self.path)

    def dump_local(self, function_output):
        import feather
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        feather.write_dataframe(function_output, self.path)





"""
Useful tools.


"""

from .dag import (
    get_etl_constraints, get_mapping_constraints, get_dm_relationship_constraints, generate_dependency_graph
    )
from .clone_tenant import (
    CloneTenant
)

from .data_model_generator import (
    DataModelGenerator
)


from unittest.mock import patch, Mock

def test_mock_password_login():
    from pycarol import PwdAuth, Carol

    mock_tenant = patch('pycarol.tenant.Tenant.get_tenant_by_domain')
    mock_get_tenant = mock_tenant.start()
    mock_get_tenant.return_value = Mock(status_code=200)
    mock_get_tenant.return_value.json.return_value = {}

    """Mocking a whole function"""
    mock_get_patcher = patch('pycarol.carol.Carol.call_api')
    response = {
        "access_token": "213133213231321231",
        "refresh_token": "asdasdasdads23123",
        "timeIssuedInMillis": 123456,
        "expires_in": 3600
    }

    mock_get = mock_get_patcher.start()
    mock_get.return_value = response

    login = Carol(domain="tenant", app_name="app", auth=PwdAuth(user="a@pycarol.com.br", password="12345"))

    assert login.auth._token.access_token is not None
    assert login.auth._token.refresh_token is not None
    assert login.auth._token.expiration is not None

    mock_get_patcher.stop()
    mock_tenant.stop()



"""Front-end for all SQL-related operations."""
from contextlib import contextmanager
import itertools
import json
import typing as T
import warnings

try:  # dataframe
    import pandas as pd
except ModuleNotFoundError:
    ...
try:  # extra
    from websocket import create_connection  # pylint: disable=import-error

    WEBSOCKET_IMPORTED = True
except ModuleNotFoundError:
    WEBSOCKET_IMPORTED = False

from pycarol import Carol, PwdAuth, PwdKeyAuth
from . import bigquery


@contextmanager
def SQLSocket(carol: Carol, *args, **kw):
    """Create Socket connection to Carol.

    Args:
        carol (Carol): Carol instance.
    """
    if not WEBSOCKET_IMPORTED:
        warnings.warn(
            """
            websocket-client is not installed.
            Please install it with:
                pip install websocket-client
        """,
            Warning,
        )
        raise ModuleNotFoundError

    if not isinstance(carol.auth, (PwdAuth, PwdKeyAuth)):
        raise TypeError("Websocket only works with `PwdAuth`, `PwdKeyAuth`")

    def _build_ws_url():
        token = carol.auth._token.access_token
        return f"wss://{carol.organization}.carol.ai/websocket/live/{token}"

    ws = create_connection(_build_ws_url())
    user_id = carol.call_api("v1/users/current")["mdmId"]
    subs = {"action": "SUBSCRIBE", "type": "SQL_RESPONSE", "subtype": user_id}
    ws.send(json.dumps(subs))
    print(json.loads(ws.recv()))
    try:
        yield ws
    finally:
        ws.close()


class SQL:
    def __init__(self, carol: Carol):
        self.carol = carol
        warnings.warn("Experimental feature. The API might change without notice.")

    def query(
        self,
        query: str,
        params: T.Optional[T.List] = None,
        method: str = "sync",
        dataframe: bool = True,
        service_account: T.Optional[T.Dict[str, str]] = None,
        dataset_id: T.Optional[str] = None,
        **kw,
    ):
        """Execute SQL Query.

        Args:
            query: `str`
                SQL Query
            params: `list` default `None`
                Params to send to Carol
            method: `str` default `sync`
                if sync runs the query synchronously and returns the result
                if socket runs the query using web socket and returns the result
                if bigquery runs the query using bigquery and returns the result
            dataframe: `bool` default `True`
                if True returns a pandas dataframe
            service_account: in case you have a service account for accessing BigQuery.
            dataset_id: overwrites the default dataset_id when using bigquery.
            **kw: `dict`
        Returns:

        """
        methods = ("sync", "socket", "bigquery")

        params = params or []
        payload = {"preparedStatement": query, "params": params}
        if method == "socket":
            results = _socket_query(self.carol, payload, **kw)
        elif method == "sync":
            results = _sync_query(self.carol, payload)
        elif method == "bigquery":
            results = bigquery.query(self.carol, query, service_account, dataset_id=dataset_id)
        else:
            raise ValueError(f"'method' must be either: {methods}")

        if method == "bigquery":
            results = [dict(row) for row in results]
        if dataframe:
            return pd.DataFrame(results)

        return results


def _socket_query(carol: Carol, payload, **kwargs):
    last_result = False
    results = []
    with SQLSocket(carol, **kwargs) as ws:
        resp = carol.call_api(path="v2/sql/query", method="POST", data=payload)
        while not last_result:
            resp = ws.recv()
            resp = json.loads(resp)
            result = json.loads(resp["message"])
            last_result = result["lastResult"]
            if result["success"]:
                results.extend(result["results"])
            else:
                raise ValueError(f"{result['error']}")
    return results


def _sync_query(carol: Carol, payload):
    results = carol.call_api(path="v2/sql/querySync", method="POST", data=payload)
    if not results[0]["success"]:
        raise ValueError(f"{results[0]['error']}")
    return list(itertools.chain(*[i["results"] for i in results]))



import re

class Online(object):
    """ Decorator to indicate the functions that will be exposed as endpoints.

    Usage:
        from pycarol.app.online import Online
        online = Online()

        @online.route("sample_endpoint")
        def sample1_function(request):
            return str(request)
    """
    def __init__(self):
        self.endpoints = {}

    def get_endpoints(self):
        return self.endpoints

    def _add_endpoint(self, name, func):
        self.endpoints[name] = func

    def _validate_route(self, name):
        if not re.match("^[a-zA-Z0-9_]*$", name):
            raise ValueError('Route name does not support special characters, only "_". Route: "{}"'.format(name))
        if len(name) > 30:
            raise ValueError('Route name should be smaller than 30 letters. Route: "{}" has {} letters'.format(name, str(len(name))))
        return True

    def route(self, name):
        def decorator(f):
            self._validate_route(name)
            self._add_endpoint(name, f)
            return f
        return decorator